"""Impala backend."""

from __future__ import annotations

import contextlib
import operator
import os
from functools import cached_property
from typing import TYPE_CHECKING, Any, Literal
from urllib.parse import parse_qs, urlparse

import impala.dbapi as impyla
import sqlglot as sg
import sqlglot.expressions as sge
from impala.error import Error as ImpylaError

import ibis.common.exceptions as com
import ibis.config
import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends.base.sqlglot import SQLGlotBackend
from ibis.backends.impala import ddl, udf
from ibis.backends.impala.client import ImpalaTable
from ibis.backends.impala.compiler import ImpalaCompiler
from ibis.backends.impala.ddl import (
    CTAS,
    CreateDatabase,
    CreateTableWithSchema,
    CreateView,
    DropDatabase,
    DropTable,
    DropView,
    RenameTable,
    TruncateTable,
)
from ibis.backends.impala.udf import (
    aggregate_function,
    scalar_function,
    wrap_uda,
    wrap_udf,
)
from ibis.config import options

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from pathlib import Path

    import pandas as pd
    import pyarrow as pa

    import ibis.expr.operations as ops


__all__ = (
    "Backend",
    "aggregate_function",
    "scalar_function",
    "wrap_uda",
    "wrap_udf",
)


class Backend(SQLGlotBackend):
    name = "impala"
    compiler = ImpalaCompiler()

    supports_in_memory_tables = True

    class Options(ibis.config.Config):
        """Impala specific options.

        Parameters
        ----------
        temp_db : str, default "__ibis_tmp"
            Database to use for temporary objects.
        temp_path : str, default "/tmp/ibis"
            Path for storage of temporary data.

        """

        temp_db: str = "__ibis_tmp"
        temp_path: str = "/tmp/__ibis"

    def _from_url(self, url: str, **kwargs: Any) -> Backend:
        """Connect to a backend using a URL `url`.

        Parameters
        ----------
        url
            URL with which to connect to a backend.
        kwargs
            Additional keyword arguments passed to the `connect` method.

        Returns
        -------
        BaseBackend
            A backend instance

        """
        url = urlparse(url)

        for name in ("username", "hostname", "port", "password"):
            if value := (
                getattr(url, name, None)
                or os.environ.get(f"{self.name.upper()}_{name.upper()}")
            ):
                kwargs[name] = value

        with contextlib.suppress(KeyError):
            kwargs["host"] = kwargs.pop("hostname")

        (database,) = url.path[1:].split("/", 1)
        if database:
            kwargs["database"] = database

        query_params = parse_qs(url.query)

        for name, value in query_params.items():
            if len(value) > 1:
                kwargs[name] = value
            elif len(value) == 1:
                kwargs[name] = value[0]
            else:
                raise com.IbisError(f"Invalid URL parameter: {name}")

        self._convert_kwargs(kwargs)
        return self.connect(**kwargs)

    def do_connect(
        self,
        host: str = "localhost",
        port: int = 21050,
        database: str = "default",
        timeout: int = 45,
        use_ssl: bool = False,
        ca_cert: str | Path | None = None,
        user: str | None = None,
        password: str | None = None,
        auth_mechanism: Literal["NOSASL", "PLAIN", "GSSAPI", "LDAP"] = "NOSASL",
        kerberos_service_name: str = "impala",
        **params: Any,
    ):
        """Create an Impala `Backend` for use with Ibis.

        Parameters
        ----------
        host
            Host name of the impalad or HiveServer2 in Hive
        port
            Impala's HiveServer2 port
        database
            Default database when obtaining new cursors
        timeout
            Connection timeout in seconds when communicating with HiveServer2
        use_ssl
            Use SSL when connecting to HiveServer2
        ca_cert
            Local path to 3rd party CA certificate or copy of server
            certificate for self-signed certificates. If SSL is enabled, but
            this argument is ``None``, then certificate validation is skipped.
        user
            LDAP user to authenticate
        password
            LDAP password to authenticate
        auth_mechanism
            |   Value    | Meaning                        |
            | :--------: | :----------------------------- |
            | `'NOSASL'` | insecure Impala connections    |
            | `'PLAIN'`  | insecure Hive clusters         |
            |  `'LDAP'`  | LDAP authenticated connections |
            | `'GSSAPI'` | Kerberos-secured clusters      |
        kerberos_service_name
            Specify a particular `impalad` service principal.
        params
            Any additional parameters necessary to open a connection to Impala.
            Please refer to impyla documentation for the full list of
            possible arguments.

        Examples
        --------
        >>> import os
        >>> import ibis
        >>> impala_host = os.environ.get("IBIS_TEST_IMPALA_HOST", "localhost")
        >>> impala_port = int(os.environ.get("IBIS_TEST_IMPALA_PORT", 21050))
        >>> client = ibis.impala.connect(host=impala_host, port=impala_port)
        >>> client  # doctest: +ELLIPSIS
        <ibis.backends.impala.Backend object at 0x...>

        """
        if ca_cert is not None:
            params["ca_cert"] = str(ca_cert)

        # make sure the connection works
        con = impyla.connect(
            host=host,
            port=port,
            database=database,
            timeout=timeout,
            use_ssl=use_ssl,
            user=user,
            password=password,
            auth_mechanism=auth_mechanism,
            kerberos_service_name=kerberos_service_name,
            **params,
        )
        with contextlib.closing(
            con.cursor(user=params.get("user"), convert_types=True)
        ) as cur:
            cur.ping()

        self.con = con
        self.options = {}
        self._ensure_temp_db_exists()

    @cached_property
    def version(self):
        with self._safe_raw_sql("SELECT VERSION()") as cursor:
            (result,) = cursor.fetchone()
        return result

    def list_databases(self, like=None):
        with self._safe_raw_sql("SHOW DATABASES") as cur:
            databases = fetchall(cur)
        return self._filter_with_like(databases.name.tolist(), like)

    def list_tables(self, like=None, database=None):
        statement = "SHOW TABLES"
        if database is not None:
            statement += f" IN {database}"

        with self._safe_raw_sql(statement) as cursor:
            tables = fetchall(cursor)
        return self._filter_with_like(tables.name.tolist(), like=like)

    def raw_sql(self, query: str):
        cursor = self.con.cursor()

        try:
            for k, v in self.options.items():
                q = f"SET {k} = {v!r}"
                util.log(q)
                cursor.execute_async(q)

            cursor._wait_to_finish()

            util.log(query)
            cursor.execute_async(query)

            cursor._wait_to_finish()
        except (Exception, KeyboardInterrupt):
            cursor.cancel_operation()
            cursor.close()
            raise

        return cursor

    def _fetch_from_cursor(self, cursor, schema):
        from ibis.formats.pandas import PandasData

        results = fetchall(cursor)
        return PandasData.convert_table(results, schema)

    @contextlib.contextmanager
    def _safe_raw_sql(self, query: str):
        if not isinstance(query, str):
            try:
                query = query.sql(dialect=self.dialect)
            except AttributeError:
                query = query.compile()

        assert isinstance(query, str), type(query)
        with contextlib.closing(self.raw_sql(query)) as cur:
            yield cur

    def _safe_exec_sql(self, *args, **kwargs):
        with self._safe_raw_sql(*args, **kwargs):
            pass

    def _fully_qualified_name(self, name, database):
        database = database or self.current_database
        return sg.table(name, db=database, quoted=self.compiler.quoted).sql(
            self.dialect
        )

    @property
    def current_database(self) -> str:
        with self._safe_raw_sql("SELECT CURRENT_DATABASE()") as cur:
            [(db,)] = cur.fetchall()
        return db

    def create_database(self, name, path=None, force=False):
        """Create a new Impala database.

        Parameters
        ----------
        name
            Database name
        path
            Path where to store the database data; otherwise uses the Impala default
        force
            Forcibly create the database

        """
        statement = CreateDatabase(name, path=path, can_exist=force)
        self._safe_exec_sql(statement)

    def drop_database(self, name, force=False):
        """Drop an Impala database.

        Parameters
        ----------
        name
            Database name
        force
            If False and there are any tables in this database, raises an
            IntegrityError

        """
        if not force or name in self.list_databases():
            tables = self.list_tables(database=name)
            udfs = self.list_udfs(database=name)
            udas = self.list_udas(database=name)
        else:
            tables = []
            udfs = []
            udas = []
        if force:
            for table in tables:
                util.log(f"Dropping {name}.{table}")
                self.drop_table_or_view(table, database=name)
            for func in udfs:
                util.log(f"Dropping function {func.name}({func.inputs})")
                self.drop_udf(
                    func.name,
                    input_types=func.inputs,
                    database=name,
                    force=True,
                )
            for func in udas:
                util.log(f"Dropping aggregate function {func.name}({func.inputs})")
                self.drop_uda(
                    func.name,
                    input_types=func.inputs,
                    database=name,
                    force=True,
                )
        elif tables or udfs or udas:
            raise com.IntegrityError(
                f"Database {name} must be empty before "
                "being dropped, or set force=True"
            )
        statement = DropDatabase(name, must_exist=not force)
        self._safe_exec_sql(statement)

    def get_schema(
        self, table_name: str, schema: str | None = None, database: str | None = None
    ) -> sch.Schema:
        """Return a Schema object for the indicated table and database.

        Parameters
        ----------
        table_name
            Table name
        schema
            Schema name. Unused in the impala backend.
        database
            Database name

        Returns
        -------
        Schema
            Ibis schema

        """
        query = sge.Describe(
            this=sg.table(
                table_name, db=schema, catalog=database, quoted=self.compiler.quoted
            )
        )

        with self._safe_raw_sql(query) as cur:
            meta = fetchall(cur)
        return sch.Schema.from_tuples(
            zip(meta["name"], meta["type"].map(self.compiler.type_mapper.from_string))
        )

    def _metadata(self, query: str) -> Iterator[tuple[str, dt.DataType]]:
        """Return a Schema object for the indicated table and database.

        Parameters
        ----------
        query
            Query to execute against Impala

        Returns
        -------
        Iterator[tuple[str, dt.DataType]]
            Iterator of column name and Ibis type pairs

        """
        tmpview = util.gen_name("impala_tmpview")
        query = f"CREATE VIEW IF NOT EXISTS {tmpview} AS {query}"

        with self._safe_raw_sql(query) as cur:
            try:
                cur.execute(f"DESCRIBE {tmpview}")
                meta = fetchall(cur)
            finally:
                cur.execute(f"DROP VIEW IF EXISTS {tmpview}")

        return zip(
            meta["name"], meta["type"].map(self.compiler.type_mapper.from_string)
        )

    @property
    def client_options(self):
        return self.con.options

    def get_options(self) -> dict[str, str]:
        """Return current query options for the Impala session."""
        with self._safe_raw_sql("SET") as cur:
            opts = fetchall(cur)

        return dict(zip(opts.option, opts.value))

    def set_options(self, options):
        self.options.update(options)

    def set_compression_codec(self, codec):
        self.set_options({"COMPRESSION_CODEC": str(codec).lower()})

    def create_view(
        self,
        name: str,
        obj: ir.Table,
        *,
        database: str | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        select = self.compile(obj)
        statement = CreateView(name, select, database=database, can_exist=overwrite)
        self._safe_exec_sql(statement)
        return self.table(name, database=database)

    def drop_view(self, name, database=None, force=False):
        stmt = DropView(name, database=database, must_exist=not force)
        self._safe_exec_sql(stmt)

    def table(self, name: str, database: str | None = None, **kwargs: Any) -> ir.Table:
        expr = super().table(name, database=database, **kwargs)
        return ImpalaTable(expr.op())

    def create_table(
        self,
        name: str,
        obj: ir.Table | None = None,
        *,
        schema=None,
        database=None,
        temp: bool | None = None,
        overwrite: bool = False,
        external: bool = False,
        format="parquet",
        location=None,
        partition=None,
        like_parquet=None,
    ) -> ir.Table:
        """Create a new table in Impala using an Ibis table expression.

        Parameters
        ----------
        name
            Table name
        obj
            If passed, creates table from select statement results
        schema
            Mutually exclusive with obj, creates an empty table with a
            particular schema
        database
            Database name
        temp
            Whether a table is temporary
        overwrite
            Do not create table if table with indicated name already exists
        external
            Create an external table; Impala will not delete the underlying
            data when the table is dropped
        format
            File format
        location
            Specify the directory location where Impala reads and writes files
            for the table
        partition
            Must pass a schema to use this. Cannot partition from an
            expression.
        like_parquet
            Can specify instead of a schema

        """
        if obj is None and schema is None:
            raise com.IbisError("The schema or obj parameter is required")

        if temp is not None:
            raise NotImplementedError(
                "Impala backend does not yet support temporary tables"
            )
        if like_parquet is not None:
            raise NotImplementedError

        if obj is not None:
            if not isinstance(obj, ir.Table):
                obj = ibis.memtable(obj)

            self._run_pre_execute_hooks(obj)

            select = self.compile(obj)

            if overwrite:
                self.drop_table(name, force=True)

            self._safe_exec_sql(
                CTAS(
                    name,
                    select,
                    database=database or self.current_database,
                    format=format,
                    external=True if location is not None else external,
                    partition=partition,
                    path=location,
                )
            )
        else:  # schema is not None
            if overwrite:
                self.drop_table(name, force=True)
            self._safe_exec_sql(
                CreateTableWithSchema(
                    name,
                    schema if schema is not None else obj.schema(),
                    database=database or self.current_database,
                    format=format,
                    external=external,
                    path=location,
                    partition=partition,
                )
            )
        return self.table(name, database=database or self.current_database)

    def avro_file(
        self, directory, avro_schema, name=None, database=None, external=True
    ):
        """Create a table to read a collection of Avro data.

        Parameters
        ----------
        directory
            Server path to directory containing avro files
        avro_schema
            The Avro schema for the data as a Python dict
        name
            Table name
        database
            Database name
        external
            Whether the table is external

        Returns
        -------
        ImpalaTable
            Impala table expression

        """
        name, database = self._get_concrete_table_path(name, database)

        stmt = ddl.CreateTableAvro(
            name, directory, avro_schema, database=database, external=external
        )
        self._safe_exec_sql(stmt)
        return self._wrap_new_table(name, database)

    def delimited_file(
        self,
        directory,
        schema,
        name=None,
        database=None,
        delimiter=",",
        na_rep=None,
        escapechar=None,
        lineterminator=None,
        external=True,
    ):
        """Interpret delimited text files as an Ibis table expression.

        See the `parquet_file` method for more details on what happens under
        the hood.

        Parameters
        ----------
        directory
            Server directory containing delimited text files
        schema
            Ibis schema
        name
            Name for the table; otherwise random names are generated
        database
            Database to create the table in
        delimiter
            Character used to delimit columns
        na_rep
            Character used to represent NULL values
        escapechar
            Character used to escape special characters
        lineterminator
            Character used to delimit lines
        external
            Create table as EXTERNAL (data will not be deleted on drop).

        Returns
        -------
        ImpalaTable
            Impala table expression

        """
        name, database = self._get_concrete_table_path(name, database)

        stmt = ddl.CreateTableDelimited(
            name,
            directory,
            schema,
            database=database,
            delimiter=delimiter,
            external=external,
            na_rep=na_rep,
            lineterminator=lineterminator,
            escapechar=escapechar,
        )
        self._safe_exec_sql(stmt)
        return self._wrap_new_table(name, database)

    def parquet_file(
        self,
        directory: str | Path,
        schema: sch.Schema | None = None,
        name: str | None = None,
        database: str | None = None,
        external: bool = True,
        like_file: str | Path | None = None,
        like_table: str | None = None,
    ):
        """Create an Ibis table from the passed directory of Parquet files.

        The table can be optionally named, otherwise a unique name will be
        generated.

        Parameters
        ----------
        directory
            Path
        schema
            If no schema provided, and neither of the like_* argument is
            passed, one will be inferred from one of the parquet files in the
            directory.
        like_file
            Absolute path to Parquet file on the server to use for schema
            definitions. An alternative to having to supply an explicit schema
        like_table
            Fully scoped and escaped string to an Impala table whose schema we
            will use for the newly created table.
        name
            Random unique name generated otherwise
        database
            Database to create the (possibly temporary) table in
        external
            If a table is external, the referenced data will not be deleted
            when the table is dropped in Impala. Otherwise Impala takes
            ownership of the Parquet file.

        Returns
        -------
        ImpalaTable
            Impala table expression

        """
        name, database = self._get_concrete_table_path(name, database)

        stmt = ddl.CreateTableParquet(
            name,
            directory,
            schema=schema,
            database=database,
            example_file=like_file,
            example_table=like_table,
            external=external,
            can_exist=False,
        )
        self._safe_exec_sql(stmt)
        return self._wrap_new_table(name, database)

    def _get_concrete_table_path(
        self, name: str | None, database: str | None
    ) -> tuple[str, str | None]:
        return name if name is not None else util.gen_name("impala_table"), database

    def _ensure_temp_db_exists(self):
        # TODO: session memoize to avoid unnecessary `SHOW DATABASES` calls
        name, path = options.impala.temp_db, options.impala.temp_path
        if name not in self.list_databases():
            self.create_database(name, path=path, force=True)

    def _drop_table(self, name: str) -> None:
        # database might have been dropped, so we suppress the
        # corresponding Exception
        with contextlib.suppress(ImpylaError):
            self.drop_table(name)

    def _wrap_new_table(self, name, database):
        qualified_name = self._fully_qualified_name(name, database)
        t = self.table(name, database=database)

        # Compute number of rows in table for better default query planning
        cardinality = t.count().execute()
        self._safe_exec_sql(
            f"ALTER TABLE {qualified_name} SET tblproperties('numRows'='{cardinality:d}', "
            "'STATS_GENERATED_VIA_STATS_TASK' = 'true')"
        )

        return t

    def insert(
        self,
        table_name,
        obj=None,
        database=None,
        overwrite=False,
        partition=None,
        values=None,
        validate=True,
    ):
        """Insert data into an existing table.

        See
        [`ImpalaTable.insert`](../backends/impala.qmd#ibis.backends.impala.client.ImpalaTable.insert)
        for parameters.

        Examples
        --------
        >>> table = "my_table"
        >>> con.insert(table, table_expr)  # quartodoc: +SKIP # doctest: +SKIP

        Completely overwrite contents
        >>> con.insert(table, table_expr, overwrite=True)  # quartodoc: +SKIP # doctest: +SKIP

        """
        if isinstance(obj, ir.Table):
            self._run_pre_execute_hooks(obj)
        table = self.table(table_name, database=database)
        return table.insert(
            obj=obj,
            overwrite=overwrite,
            partition=partition,
            values=values,
            validate=validate,
        )

    def drop_table(
        self, name: str, *, database: str | None = None, force: bool = False
    ) -> None:
        """Drop an Impala table.

        Parameters
        ----------
        name
            Table name
        database
            Database name
        force
            Database may throw exception if table does not exist

        Examples
        --------
        >>> table = "my_table"
        >>> db = "operations"
        >>> con.drop_table(table, database=db, force=True)  # quartodoc: +SKIP # doctest: +SKIP

        """
        statement = DropTable(name, database=database, must_exist=not force)
        self._safe_exec_sql(statement)

    def truncate_table(self, name: str, database: str | None = None) -> None:
        """Delete all rows from an existing table.

        Parameters
        ----------
        name
            Table name
        database
            Database name

        """
        statement = TruncateTable(name, database=database)
        self._safe_exec_sql(statement)

    def rename_table(self, old_name: str, new_name: str) -> None:
        """Rename an existing table.

        Parameters
        ----------
        old_name
            The old name of the table.
        new_name
            The new name of the table.

        """
        statement = RenameTable(old_name, new_name)
        self._safe_exec_sql(statement)

    def drop_table_or_view(self, name, *, database=None, force=False):
        """Drop view or table."""
        try:
            self.drop_table(name, database=database)
        except Exception as e:  # noqa: BLE001
            try:
                self.drop_view(name, database=database)
            except Exception:  # noqa: BLE001
                raise e

    def cache_table(self, table_name, *, database=None, pool="default"):
        """Caches a table in cluster memory in the given pool.

        Parameters
        ----------
        table_name
            Table name
        database
            Database name
        pool
            The name of the pool in which to cache the table

        Examples
        --------
        >>> table = "my_table"
        >>> db = "operations"
        >>> pool = "op_4GB_pool"
        >>> con.cache_table("my_table", database=db, pool=pool)  # quartodoc: +SKIP # doctest: +SKIP

        """
        statement = ddl.CacheTable(table_name, database=database, pool=pool)
        self._safe_exec_sql(statement)

    def create_function(self, func, name=None, database=None):
        """Create a function within Impala.

        Parameters
        ----------
        func
            UDF or UDAF
        name
            Function name
        database
            Database name

        """
        if name is None:
            name = func.name
        database = database or self.current_database

        if isinstance(func, udf.ImpalaUDF):
            stmt = ddl.CreateUDF(func, name=name, database=database)
        elif isinstance(func, udf.ImpalaUDA):
            stmt = ddl.CreateUDA(func, name=name, database=database)
        else:
            raise TypeError(func)
        self._safe_exec_sql(stmt)

    def drop_udf(
        self,
        name,
        input_types=None,
        database=None,
        force=False,
        aggregate=False,
    ):
        """Drop a UDF.

        If only name is given, this will search for the relevant UDF and drop
        it. To delete an overloaded UDF, give only a name and force=True

        Parameters
        ----------
        name
            Function name
        input_types
            Input types
        force
            Must be set to `True` to drop overloaded UDFs
        database
            Database name
        aggregate
            Whether the function is an aggregate

        """
        if not input_types:
            if not database:
                database = self.current_database
            result = self.list_udfs(database=database, like=name)
            if len(result) > 1:
                if force:
                    for func in result:
                        self._drop_single_function(
                            func.name,
                            func.inputs,
                            database=database,
                            aggregate=aggregate,
                        )
                    return
                else:
                    raise com.DuplicateUDFError(name)
            elif len(result) == 1:
                func = result.pop()
                self._drop_single_function(
                    func.name,
                    func.inputs,
                    database=database,
                    aggregate=aggregate,
                )
                return
            else:
                raise com.MissingUDFError(name)
        self._drop_single_function(
            name, input_types, database=database, aggregate=aggregate
        )

    def drop_uda(self, name, input_types=None, database=None, force=False):
        """Drop an aggregate function."""
        return self.drop_udf(
            name, input_types=input_types, database=database, force=force
        )

    def _drop_single_function(self, name, input_types, database=None, aggregate=False):
        stmt = ddl.DropFunction(
            name,
            input_types,
            must_exist=False,
            aggregate=aggregate,
            database=database,
        )
        self._safe_exec_sql(stmt)

    def _drop_all_functions(self, database):
        udfs = self.list_udfs(database=database)
        for fnct in udfs:
            stmt = ddl.DropFunction(
                fnct.name,
                fnct.inputs,
                must_exist=False,
                aggregate=False,
                database=database,
            )
            self._safe_exec_sql(stmt)
        udafs = self.list_udas(database=database)
        for udaf in udafs:
            stmt = ddl.DropFunction(
                udaf.name,
                udaf.inputs,
                must_exist=False,
                aggregate=True,
                database=database,
            )
            self._safe_exec_sql(stmt)

    def list_udfs(self, database=None, like=None):
        """Lists all UDFs associated with given database."""
        if not database:
            database = self.current_database
        statement = ddl.ListFunction(database, like=like, aggregate=False)
        with self._safe_raw_sql(statement) as cur:
            return self._get_udfs(cur)

    def list_udas(self, database=None, like=None):
        """Lists all UDAFs associated with a given database."""
        if not database:
            database = self.current_database
        statement = ddl.ListFunction(database, like=like, aggregate=True)
        with self._safe_raw_sql(statement) as cur:
            return self._get_udfs(cur)

    def _get_udfs(self, cur):
        rows = fetchall(cur)

        if rows.empty:
            return []

        current_database = self.current_database
        type_mapper = self.compiler.type_mapper
        result = []
        for return_type, signature, *_ in rows.itertuples(index=False):
            anon = sg.parse_one(signature)
            name = anon.this
            inputs = [
                type_mapper.from_string(expr.this.this) for expr in anon.expressions
            ]

            output = type_mapper.from_string(return_type)

            result.append((current_database, name, tuple(inputs), output))
        return result

    def exists_udf(self, name: str, database: str | None = None) -> bool:
        """Checks if a given UDF exists within a specified database."""
        return bool(self.list_udfs(database=database, like=name))

    def exists_uda(self, name: str, database: str | None = None) -> bool:
        """Checks if a given UDAF exists within a specified database."""
        return bool(self.list_udas(database=database, like=name))

    def compute_stats(
        self,
        name: str,
        database: str | None = None,
        incremental: bool = False,
    ) -> None:
        """Issue a `COMPUTE STATS` command for a given table.

        Parameters
        ----------
        name
            Can be fully qualified (with database name)
        database
            Database name
        incremental
            If True, issue COMPUTE INCREMENTAL STATS

        """
        maybe_inc = "INCREMENTAL " if incremental else ""
        cmd = f"COMPUTE {maybe_inc}STATS"

        stmt = self._table_command(cmd, name, database=database)
        self._safe_exec_sql(stmt)

    def invalidate_metadata(
        self,
        name: str | None = None,
        database: str | None = None,
    ) -> None:
        """Issue an `INVALIDATE METADATA` command.

        Optionally this applies to a specific table. See Impala documentation.

        Parameters
        ----------
        name
            Table name. Can be fully qualified (with database)
        database
            Database name

        """
        stmt = "INVALIDATE METADATA"
        if name is not None:
            stmt = self._table_command(stmt, name, database=database)
        self._safe_exec_sql(stmt)

    def refresh(self, name: str, database: str | None = None) -> None:
        """Reload metadata for a table.

        This can be useful after ingesting data as part of an ETL pipeline, for
        example.

        Related to `INVALIDATE METADATA`. See Impala documentation for more.

        Parameters
        ----------
        name
            Table name. Can be fully qualified (with database)
        database
            Database name

        """
        # TODO(wesm): can this statement be cancelled?
        stmt = self._table_command("REFRESH", name, database=database)
        self._safe_exec_sql(stmt)

    def describe_formatted(
        self,
        name: str,
        database: str | None = None,
    ) -> pd.DataFrame:
        """Retrieve the results of a `DESCRIBE FORMATTED` command.

        See Impala documentation for more.

        Parameters
        ----------
        name
            Table name. Can be fully qualified (with database)
        database
            Database name

        """
        from ibis.backends.impala.metadata import parse_metadata

        stmt = self._table_command("DESCRIBE FORMATTED", name, database=database)
        result = self._exec_statement(stmt)

        # Leave formatting to pandas
        for c in result.columns:
            result[c] = result[c].str.strip()

        return parse_metadata(result)

    def show_files(
        self,
        name: str,
        database: str | None = None,
    ) -> pd.DataFrame:
        """Retrieve results of a `SHOW FILES` command for a table.

        See Impala documentation for more.

        Parameters
        ----------
        name
            Table name. Can be fully qualified (with database)
        database
            Database name

        """
        stmt = self._table_command("SHOW FILES IN", name, database=database)
        return self._exec_statement(stmt)

    def list_partitions(self, name, database=None):
        stmt = self._table_command("SHOW PARTITIONS", name, database=database)
        return self._exec_statement(stmt)

    def table_stats(self, name, database=None):
        """Return results of `SHOW TABLE STATS` for the table `name`."""
        stmt = self._table_command("SHOW TABLE STATS", name, database=database)
        return self._exec_statement(stmt)

    def column_stats(self, name, database=None):
        """Return results of `SHOW COLUMN STATS` for the table `name`."""
        stmt = self._table_command("SHOW COLUMN STATS", name, database=database)
        return self._exec_statement(stmt)

    def _exec_statement(self, stmt):
        with self._safe_raw_sql(stmt) as cur:
            return fetchall(cur)

    def _table_command(self, cmd, name, database=None):
        qualified_name = self._fully_qualified_name(name, database)
        return f"{cmd} {qualified_name}"

    def to_pyarrow(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pa.Table:
        import pyarrow as pa
        import pyarrow_hotfix  # noqa: F401

        from ibis.formats.pyarrow import PyArrowData

        self._run_pre_execute_hooks(expr)

        table_expr = expr.as_table()
        output = pa.Table.from_pandas(
            self.execute(table_expr, params=params, limit=limit, **kwargs),
            preserve_index=False,
        )
        table = PyArrowData.convert_table(output, table_expr.schema())
        return expr.__pyarrow_result__(table)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1000000,
        **kwargs: Any,
    ) -> pa.ipc.RecordBatchReader:
        pa = self._import_pyarrow()
        self._run_pre_execute_hooks(expr)

        pa_table = self.to_pyarrow(
            expr.as_table(), params=params, limit=limit, **kwargs
        )
        return pa.RecordBatchReader.from_batches(
            pa_table.schema, pa_table.to_batches(max_chunksize=chunk_size)
        )

    def explain(
        self, expr: ir.Expr | str, params: Mapping[ir.Expr, Any] | None = None
    ) -> str:
        """Explain an expression.

        Return the query plan associated with the indicated expression or SQL
        query.

        Returns
        -------
        str
            Query plan

        """
        query = self.compile(expr, params=params)
        statement = f"EXPLAIN {query}"

        with self._safe_raw_sql(statement) as cur:
            results = fetchall(cur)

        return "\n".join(["Query:", util.indent(query, 2), "", *results.iloc[:, 0]])

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema
        if null_columns := [col for col, dtype in schema.items() if dtype.is_null()]:
            raise com.IbisTypeError(
                "Impala cannot yet reliably handle `null` typed columns; "
                f"got null typed columns: {null_columns}"
            )

        # only register if we haven't already done so
        if (name := op.name) not in self.list_tables():
            type_mapper = self.compiler.type_mapper
            quoted = self.compiler.quoted
            column_defs = [
                sg.exp.ColumnDef(
                    this=sg.to_identifier(colname, quoted=quoted),
                    kind=type_mapper.from_ibis(typ),
                    # we don't support `NOT NULL` constraints in trino because
                    # because each trino connector differs in whether it
                    # supports nullability constraints, and whether the
                    # connector supports it isn't visible to ibis via a
                    # metadata query
                )
                for colname, typ in schema.items()
            ]

            create_stmt = sg.exp.Create(
                kind="TABLE",
                this=sg.exp.Schema(
                    this=sg.to_identifier(name, quoted=quoted), expressions=column_defs
                ),
            ).sql(self.name, pretty=True)

            data = op.data.to_frame().itertuples(index=False)
            specs = ", ".join("?" * len(schema))
            table = sg.table(name, quoted=quoted).sql(self.name)
            insert_stmt = f"INSERT INTO {table} VALUES ({specs})"
            with self._safe_raw_sql(create_stmt) as cur:
                for row in data:
                    cur.execute(insert_stmt, row)


def fetchall(cur):
    batches = cur.fetchcolumnar()
    names = list(map(operator.itemgetter(0), cur.description))
    df = _column_batches_to_dataframe(names, batches)
    return df


def _column_batches_to_dataframe(names, batches):
    import pandas as pd

    cols = {
        name: _chunks_to_pandas_array(chunks)
        for name, chunks in zip(names, zip(*(b.columns for b in batches)))
    }

    return pd.DataFrame(cols, columns=names)


_HS2_TTypeId_to_dtype = {
    "BOOLEAN": "bool",
    "TINYINT": "int8",
    "SMALLINT": "int16",
    "INT": "int32",
    "BIGINT": "int64",
    "TIMESTAMP": "datetime64[ns]",
    "FLOAT": "float32",
    "DOUBLE": "float64",
    "STRING": "object",
    "DECIMAL": "object",
    "BINARY": "object",
    "VARCHAR": "object",
    "CHAR": "object",
    "DATE": "datetime64[ns]",
    "VOID": None,
}


def _chunks_to_pandas_array(chunks):
    import numpy as np

    total_length = 0
    have_nulls = False
    for c in chunks:
        total_length += len(c)
        have_nulls = have_nulls or c.nulls.any()

    type_ = chunks[0].data_type
    numpy_type = _HS2_TTypeId_to_dtype[type_]

    def fill_nonnull(target, chunks):
        pos = 0
        for c in chunks:
            target[pos : pos + len(c)] = c.values
            pos += len(c.values)

    def fill(target, chunks, na_rep):
        pos = 0
        for c in chunks:
            nulls = c.nulls.copy()
            nulls.bytereverse()
            bits = np.frombuffer(nulls.tobytes(), dtype="u1")
            mask = np.unpackbits(bits).view(np.bool_)

            k = len(c)

            dest = target[pos : pos + k]
            dest[:] = c.values
            dest[mask[:k]] = na_rep

            pos += k

    if have_nulls:
        if numpy_type in ("bool", "datetime64[ns]"):
            target = np.empty(total_length, dtype="O")
            na_rep = np.nan
        elif numpy_type.startswith("int"):
            target = np.empty(total_length, dtype="f8")
            na_rep = np.nan
        else:
            target = np.empty(total_length, dtype=numpy_type)
            na_rep = np.nan

        fill(target, chunks, na_rep)
    else:
        target = np.empty(total_length, dtype=numpy_type)
        fill_nonnull(target, chunks)

    return target



from __future__ import annotations

import contextlib

import pandas as pd
import pytest
import sqlglot as sg
from pytest import param

import ibis
import ibis.common.exceptions as com
from ibis import _
from ibis.backends.base import _get_backend_names

# import here to load the dialect in to sqlglot so we can use it for transpilation
from ibis.backends.base.sqlglot.dialects import (  # noqa: F401
    MSSQL,
    DataFusion,
    Druid,
    Exasol,
    Flink,
    Impala,
    Polars,
    PySpark,
    RisingWave,
)
from ibis.backends.tests.errors import (
    ExaQueryError,
    GoogleBadRequest,
    OracleDatabaseError,
    PolarsComputeError,
)

dot_sql_never = pytest.mark.never(
    ["dask", "pandas"], reason="dask and pandas do not accept SQL"
)

_NAMES = {
    "bigquery": "ibis_gbq_testing.functional_alltypes",
    "exasol": '"functional_alltypes"',
}


@pytest.mark.notyet(["oracle"], reason="table quoting behavior")
@dot_sql_never
@pytest.mark.parametrize(
    "schema",
    [
        param(None, id="implicit_schema", marks=[pytest.mark.notimpl(["druid"])]),
        param({"s": "string", "new_col": "double"}, id="explicit_schema"),
    ],
)
def test_con_dot_sql(backend, con, schema):
    alltypes = backend.functional_alltypes
    # pull out the quoted name
    name = _NAMES.get(con.name, "functional_alltypes")
    quoted = True
    cols = [
        sg.column("string_col", quoted=quoted).as_("s", quoted=quoted).sql(con.dialect),
        (sg.column("double_col", quoted=quoted) + 1.0)
        .as_("new_col", quoted=quoted)
        .sql(con.dialect),
    ]
    t = (
        con.sql(
            f"SELECT {', '.join(cols)} FROM {name}",
            schema=schema,
        )
        .group_by("s")  # group by a column from SQL
        .aggregate(yas=lambda t: t.new_col.max())
        .order_by("yas")
    )

    alltypes_df = alltypes.execute()
    result = t.execute()["yas"]
    expected = (
        alltypes_df.assign(
            s=alltypes_df.string_col, new_col=alltypes_df.double_col + 1.0
        )
        .groupby("s")
        .new_col.max()
        .rename("yas")
        .sort_values()
        .reset_index(drop=True)
    )
    backend.assert_series_equal(result.astype(expected.dtype), expected)


@pytest.mark.notyet(["polars"], raises=PolarsComputeError)
@pytest.mark.notyet(
    ["bigquery"], raises=GoogleBadRequest, reason="requires a qualified name"
)
@pytest.mark.notyet(
    ["druid"], raises=com.IbisTypeError, reason="druid does not preserve case"
)
@dot_sql_never
def test_table_dot_sql(backend):
    alltypes = backend.functional_alltypes
    t = (
        alltypes.sql(
            """
            SELECT
              "string_col" AS "s",
              "double_col" + CAST(1.0 AS DOUBLE) AS "new_col"
            FROM "functional_alltypes"
            """,
            dialect="duckdb",
        )
        .group_by("s")  # group by a column from SQL
        .aggregate(fancy_af=lambda t: t.new_col.mean())
        .alias("awesome_t")  # create a name for the aggregate
        .sql('SELECT "fancy_af" AS "yas" FROM "awesome_t"', dialect="duckdb")
        .order_by(_.yas)
    )

    alltypes_df = alltypes.execute()
    result = t.execute()["yas"]
    expected = (
        alltypes_df.assign(
            s=alltypes_df.string_col, new_col=alltypes_df.double_col + 1.0
        )
        .groupby("s")
        .new_col.mean()
        .rename("yas")
        .reset_index()
        .yas
    )
    assert pytest.approx(result) == expected


@pytest.mark.notyet(["polars"], raises=PolarsComputeError)
@pytest.mark.notyet(
    ["bigquery"], raises=GoogleBadRequest, reason="requires a qualified name"
)
@pytest.mark.notyet(
    ["druid"], raises=com.IbisTypeError, reason="druid does not preserve case"
)
@pytest.mark.notimpl(
    ["oracle"],
    OracleDatabaseError,
    reason="oracle doesn't know which of the tables in the join to sort from",
)
@dot_sql_never
def test_table_dot_sql_with_join(backend):
    alltypes = backend.functional_alltypes
    t = (
        alltypes.sql(
            """
            SELECT
              "string_col" AS "s",
              "double_col" + CAST(1.0 AS DOUBLE) AS "new_col"
            FROM "functional_alltypes"
            """,
            dialect="duckdb",
        )
        .alias("ft")
        .group_by("s")  # group by a column from SQL
        .aggregate(fancy_af=lambda t: t.new_col.mean())
        .alias("awesome_t")  # create a name for the aggregate
        .sql(
            """
            SELECT
              "l"."fancy_af" AS "yas",
              "r"."s" AS "s"
            FROM "awesome_t" AS "l"
            LEFT JOIN "ft" AS "r"
            ON "l"."s" = "r"."s"
            """,  # clickhouse needs the r.s AS s, otherwise the column name is returned as r.s
            dialect="duckdb",
        )
        .order_by(["s", "yas"])
    )

    alltypes_df = alltypes.execute()
    result = t.execute()

    ft = alltypes_df.assign(
        s=alltypes_df.string_col, new_col=alltypes_df.double_col + 1.0
    )
    expected = pd.merge(
        ft.groupby("s").new_col.mean().rename("yas").reset_index(),
        ft[["s"]],
        on=["s"],
        how="left",
    )[["yas", "s"]].sort_values(["s", "yas"])
    backend.assert_frame_equal(result, expected)


@pytest.mark.notyet(["polars"], raises=PolarsComputeError)
@pytest.mark.notyet(["druid"], reason="druid doesn't respect column name case")
@pytest.mark.notyet(
    ["bigquery"], raises=GoogleBadRequest, reason="requires a qualified name"
)
@dot_sql_never
def test_table_dot_sql_repr(backend):
    alltypes = backend.functional_alltypes
    t = (
        alltypes.sql(
            """
            SELECT
              "string_col" AS "s",
              "double_col" + CAST(1.0 AS DOUBLE) AS "new_col"
            FROM "functional_alltypes"
            """,
            dialect="duckdb",
        )
        .group_by("s")  # group by a column from SQL
        .aggregate(fancy_af=lambda t: t.new_col.mean())
        .alias("awesome_t")  # create a name for the aggregate
        .sql(
            'SELECT "fancy_af" AS "yas" FROM "awesome_t" ORDER BY "fancy_af"',
            dialect="duckdb",
        )
    )

    assert repr(t)


@dot_sql_never
def test_dot_sql_alias_with_params(backend, alltypes, df):
    t = alltypes
    x = t.select(x=t.string_col + " abc").alias("foo")
    result = x.execute()
    expected = df.string_col.add(" abc").rename("x")
    backend.assert_series_equal(result.x, expected)


@dot_sql_never
def test_dot_sql_reuse_alias_with_different_types(backend, alltypes, df):
    foo1 = alltypes.select(x=alltypes.string_col).alias("foo")
    foo2 = alltypes.select(x=alltypes.bigint_col).alias("foo")
    expected1 = df.string_col.rename("x")
    expected2 = df.bigint_col.rename("x")
    backend.assert_series_equal(foo1.x.execute(), expected1)
    backend.assert_series_equal(foo2.x.execute(), expected2)


_NO_SQLGLOT_DIALECT = ("pandas", "dask")
no_sqlglot_dialect = [
    param(dialect, marks=pytest.mark.xfail) for dialect in sorted(_NO_SQLGLOT_DIALECT)
]
dialects = sorted(_get_backend_names(exclude=_NO_SQLGLOT_DIALECT)) + no_sqlglot_dialect


@pytest.mark.parametrize("dialect", dialects)
@pytest.mark.notyet(["polars"], raises=PolarsComputeError)
@dot_sql_never
@pytest.mark.notyet(["druid"], reason="druid doesn't respect column name case")
def test_table_dot_sql_transpile(backend, alltypes, dialect, df):
    name = "foo2"
    foo = alltypes.select(x=_.bigint_col + 1).alias(name)
    expr = sg.select(sg.column("x", quoted=True)).from_(sg.table(name, quoted=True))
    sqlstr = expr.sql(dialect=dialect, pretty=True)
    dot_sql_expr = foo.sql(sqlstr, dialect=dialect)
    result = dot_sql_expr.execute()
    expected = df.bigint_col.add(1).rename("x")
    backend.assert_series_equal(result.x, expected)


@pytest.mark.parametrize("dialect", dialects)
@pytest.mark.notyet(
    ["druid"], raises=AttributeError, reason="druid doesn't respect column names"
)
@pytest.mark.notyet(["bigquery"])
@dot_sql_never
def test_con_dot_sql_transpile(backend, con, dialect, df):
    t = sg.table("functional_alltypes", quoted=True)
    foo = sg.select(
        sg.alias(sg.column("bigint_col", quoted=True) + 1, "x", quoted=True)
    ).from_(t)
    sqlstr = foo.sql(dialect=dialect, pretty=True)
    expr = con.sql(sqlstr, dialect=dialect)
    result = expr.execute()
    expected = df.bigint_col.add(1).rename("x")
    backend.assert_series_equal(result.x, expected)


@dot_sql_never
@pytest.mark.notimpl(["druid", "polars"])
@pytest.mark.notimpl(
    ["exasol"],
    raises=ExaQueryError,
    reason="loading the test data is a pain because of embedded commas",
)
def test_order_by_no_projection(backend):
    con = backend.connection
    expr = (
        backend.astronauts.group_by("name")
        .agg(nbr_missions=_.count())
        .order_by(_.nbr_missions.desc())
    )

    result = con.sql(ibis.to_sql(expr)).execute().name.iloc[:2]
    assert set(result) == {"Ross, Jerry L.", "Chang-Diaz, Franklin R."}


@dot_sql_never
@pytest.mark.notyet(["polars"], raises=PolarsComputeError)
def test_dot_sql_limit(con):
    expr = con.sql('SELECT * FROM (SELECT \'abc\' "ts") "x"', dialect="duckdb").limit(1)
    result = expr.execute()

    assert len(result) == 1
    assert len(result.columns) == 1
    assert result.columns[0].lower() == "ts"
    assert result.iat[0, 0] == "abc"


@pytest.fixture(scope="module")
def mem_t(con):
    if con.name == "druid":
        pytest.xfail("druid does not support create_table")

    name = ibis.util.gen_name(con.name)

    # flink only supports memtables if `temp` is True, seems like we should
    # address that for users
    con.create_table(
        name, ibis.memtable({"a": list("def")}), temp=con.name == "flink" or None
    )
    yield name
    with contextlib.suppress(NotImplementedError):
        con.drop_table(name, force=True)


@dot_sql_never
@pytest.mark.notyet(["polars"], raises=NotImplementedError)
def test_cte(con, mem_t):
    t = con.table(mem_t)
    foo = t.alias("foo")
    assert foo.schema() == t.schema()
    assert foo.count().execute() == t.count().execute()

    expr = foo.sql('SELECT count(*) "x" FROM "foo"', dialect="duckdb")
    assert expr.execute().iat[0, 0] == t.count().execute()



from __future__ import annotations

import itertools
import math
from typing import Callable

import numpy as np
import pandas as pd

from ibis.util import gen_name


def isnull(obj):
    return obj is None or obj is pd.NA or (isinstance(obj, float) and math.isnan(obj))


class PandasUtils:
    @classmethod
    def merge(cls, *args, **kwargs):
        return pd.merge(*args, **kwargs)

    @classmethod
    def merge_asof(cls, *args, **kwargs):
        return pd.merge_asof(*args, **kwargs)

    @classmethod
    def concat(cls, dfs, **kwargs):
        return pd.concat(dfs, **kwargs)

    @classmethod
    def asseries(cls, value, like=None):
        """Ensure that value is a pandas Series object, broadcast if necessary."""
        size = len(like) if like is not None else 1
        if isinstance(value, pd.Series):
            return value
        elif isinstance(value, (list, np.ndarray)):
            return pd.Series(itertools.repeat(np.array(value), size))
        else:
            return pd.Series(np.repeat(value, size))

    @classmethod
    def asframe(cls, values: dict | tuple, concat=True):
        """Construct a DataFrame from a dict or tuple of Series objects."""
        if isinstance(values, dict):
            names, values = zip(*values.items())
        elif isinstance(values, tuple):
            names = [f"_{i}" for i in range(len(values))]
        else:
            raise TypeError(f"values must be a dict, or tuple; got {type(values)}")

        all_scalars = True
        representative = None
        for v in values:
            if isinstance(v, pd.Series):
                representative = v
                all_scalars = False
                break

        columns = [cls.asseries(v, like=representative) for v in values]
        if concat:
            df = pd.concat(columns, axis=1, keys=names)
            return df, all_scalars
        else:
            return columns, all_scalars

    @classmethod
    def agg(cls, func, arg_column, where_column):
        if where_column is None:

            def applier(df):
                return func(df[arg_column.name])
        else:

            def applier(df):
                mask = df[where_column.name]
                col = df[arg_column.name][mask]
                return func(col)

        return applier

    @classmethod
    def generic(cls, func: Callable, operands, **kwargs):
        return func(*operands.values())

    @classmethod
    def rowwise(cls, func: Callable, operands, **kwargs):
        """Kernel applied to a row, where all the operands are scalars."""
        # dealing with a collection of series objects
        df, _ = cls.asframe(operands)
        return df.apply(func, axis=1)

    @classmethod
    def columnwise(cls, func: Callable, operands, **kwargs):
        """Kernel where all the operands are series objects."""
        df, _ = cls.asframe(operands)
        return func(df)

    @classmethod
    def serieswise(cls, func, operands, **kwargs):
        """Kernel where the first operand is a series object."""
        (key, value), *rest = operands.items()
        # ensure that the first operand is a series object
        value = cls.asseries(value)
        operands = {key: value, **dict(rest)}
        return func(**operands)

    @classmethod
    def elementwise(cls, func, operands, **kwargs):
        """Kernel applied to an element, where all the operands are scalars."""
        value = operands.pop(next(iter(operands)))
        if isinstance(value, pd.Series):
            # dealing with a single series object
            if operands:
                return value.apply(func, **operands)
            else:
                return value.map(func, na_action="ignore")
        else:
            # dealing with a single scalar object
            return func(value, **operands)


class UngroupedFrame:
    def __init__(self, df):
        self.df = df

    def groups(self):
        yield self.df

    def apply_reduction(self, func, **kwargs):
        result = func(self.df, **kwargs)
        data = [result] * len(self.df)
        return pd.Series(data, index=self.df.index)

    def apply_analytic(self, func, **kwargs):
        return func(self.df, **kwargs)


class GroupedFrame:
    def __init__(self, df, group_keys):
        self.df = df
        self.group_keys = group_keys
        self.groupby = df.groupby(group_keys, as_index=True)

    def groups(self):
        for _, df in self.groupby:
            yield df

    def apply_analytic(self, func, **kwargs):
        results = [func(df, **kwargs) for df in self.groups()]
        return pd.concat(results)

    def apply_reduction(self, func, **kwargs):
        name = gen_name("result")
        result = self.groupby.apply(func, **kwargs).rename(name)
        df = self.df.merge(result, left_on=self.group_keys, right_index=True)
        return df[name]


class RowsFrame:
    def __init__(self, parent):
        self.parent = parent

    @staticmethod
    def adjust(length, index, start_offset, end_offset):
        if start_offset is None:
            start_index = 0
        else:
            start_index = index + start_offset
            if start_index < 0:
                start_index = 0
            elif start_index > length:
                start_index = length

        if end_offset is None:
            end_index = length
        else:
            end_index = index + end_offset + 1
            if end_index < 0:
                end_index = 0
            elif end_index > length:
                end_index = length

        return (start_index, end_index)

    def apply_analytic(self, func, **kwargs):
        return self.parent.apply_analytic(func, **kwargs)

    def apply_reduction(self, func, **kwargs):
        results = {}
        for df in self.parent.groups():
            for i, (ix, row) in enumerate(df.iterrows()):
                # TODO(kszucs): use unique column names for _start, _end
                start, end = row["__start__"], row["__end__"]
                start_index, end_index = self.adjust(len(df), i, start, end)
                subdf = df.iloc[start_index:end_index]
                results[ix] = func(subdf, **kwargs)

        return pd.Series(results)


class RangeFrame:
    def __init__(self, parent, order_key):
        self.parent = parent
        self.order_key = order_key

    @staticmethod
    def predicate(col, i, start, end):
        value = col.iat[i]
        if start is None:
            return col <= value + end
        elif end is None:
            return col >= value + start
        else:
            return (col >= value + start) & (col <= value + end)

    def apply_analytic(self, func, **kwargs):
        return self.parent.apply_analytic(func, **kwargs)

    def apply_reduction(self, func, **kwargs):
        results = {}
        for df in self.parent.groups():
            for i, (ix, row) in enumerate(df.iterrows()):
                start, end = row["__start__"], row["__end__"]
                column = df[self.order_key]
                predicate = self.predicate(column, i, start, end)
                subdf = df[predicate]
                results[ix] = func(subdf, **kwargs)

        return pd.Series(results)



"""APIs for creating user-defined functions."""

from __future__ import annotations

import ibis.legacy.udf.vectorized


class udf:
    @staticmethod
    def elementwise(input_type, output_type):
        """Alias for ibis.legacy.udf.vectorized.elementwise."""

        return ibis.legacy.udf.vectorized.elementwise(input_type, output_type)

    @staticmethod
    def reduction(input_type, output_type):
        """Alias for ibis.legacy.udf.vectorized.reduction."""
        return ibis.legacy.udf.vectorized.reduction(input_type, output_type)

    @staticmethod
    def analytic(input_type, output_type):
        """Alias for ibis.legacy.udf.vectorized.analytic."""
        return ibis.legacy.udf.vectorized.analytic(input_type, output_type)



from __future__ import annotations

import operator
from functools import reduce

import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd

import ibis.backends.dask.kernels as dask_kernels
import ibis.expr.operations as ops
from ibis.backends.dask.convert import DaskConverter
from ibis.backends.dask.helpers import (
    DaskUtils,
    add_globally_consecutive_column,
)
from ibis.backends.pandas.executor import PandasExecutor
from ibis.backends.pandas.rewrites import (
    PandasAggregate,
    PandasJoin,
    PandasLimit,
    PandasResetIndex,
    PandasScalarSubquery,
    plan,
)
from ibis.common.exceptions import UnboundExpressionError
from ibis.formats.pandas import PandasData, PandasType
from ibis.util import gen_name

# ruff: noqa: F811


class DaskExecutor(PandasExecutor, DaskUtils):
    name = "dask"
    kernels = dask_kernels

    @classmethod
    def visit(cls, op: ops.Node, **kwargs):
        return super().visit(op, **kwargs)

    @classmethod
    def visit(cls, op: ops.Cast, arg, to):
        if arg is None:
            return None
        elif isinstance(arg, dd.Series):
            return DaskConverter.convert_column(arg, to)
        else:
            return DaskConverter.convert_scalar(arg, to)

    @classmethod
    def visit(
        cls, op: ops.SimpleCase | ops.SearchedCase, cases, results, default, base=None
    ):
        def mapper(df, cases, results, default):
            cases = [case.astype("bool") for case in cases]
            cases.append(pd.Series(True, index=df.index))

            results.append(default)
            out = np.select(cases, results)

            return pd.Series(out, index=df.index)

        dtype = PandasType.from_ibis(op.dtype)
        if base is not None:
            cases = tuple(base == case for case in cases)
        kwargs = dict(cases=cases, results=results, default=default)

        return cls.partitionwise(mapper, kwargs, name=op.name, dtype=dtype)

    @classmethod
    def visit(cls, op: ops.TimestampTruncate | ops.DateTruncate, arg, unit):
        # TODO(kszucs): should use serieswise()
        unit = {"m": "Min", "ms": "L"}.get(unit.short, unit.short)
        try:
            return arg.dt.floor(unit)
        except ValueError:
            return arg.dt.to_period(unit).dt.to_timestamp()

    @classmethod
    def visit(cls, op: ops.IntervalFromInteger, unit, **kwargs):
        if unit.short in {"Y", "Q", "M", "W"}:
            return cls.elementwise(
                lambda v: pd.DateOffset(**{unit.plural: v}),
                kwargs,
                name=op.name,
                dtype=object,
            )
        else:
            return cls.serieswise(
                lambda arg: arg.astype(f"timedelta64[{unit.short}]"), kwargs
            )

    @classmethod
    def visit(cls, op: ops.BetweenTime, arg, lower_bound, upper_bound):
        if getattr(arg.dtype, "tz", None) is not None:
            localized = arg.dt.tz_convert("UTC").dt.tz_localize(None)
        else:
            localized = arg

        time = localized.dt.time.astype(str)
        indexer = ((time >= lower_bound) & (time <= upper_bound)).to_dask_array(True)

        result = da.zeros(len(arg), dtype=np.bool_)
        result[indexer] = True
        return dd.from_array(result)

    @classmethod
    def visit(cls, op: ops.FindInSet, needle, values):
        def mapper(df, cases):
            thens = [i for i, _ in enumerate(cases)]
            out = np.select(cases, thens, default=-1)
            return pd.Series(out, index=df.index)

        dtype = PandasType.from_ibis(op.dtype)
        cases = [needle == value for value in values]
        kwargs = dict(cases=cases)
        return cls.partitionwise(mapper, kwargs, name=op.name, dtype=dtype)

    @classmethod
    def visit(cls, op: ops.Array, exprs):
        return cls.rowwise(
            lambda row: np.array(row, dtype=object), exprs, name=op.name, dtype=object
        )

    @classmethod
    def visit(cls, op: ops.ArrayConcat, arg):
        dtype = PandasType.from_ibis(op.dtype)
        return cls.rowwise(
            lambda row: np.concatenate(row.values), arg, name=op.name, dtype=dtype
        )

    @classmethod
    def visit(cls, op: ops.Unnest, arg):
        arg = cls.asseries(arg)
        mask = arg.map(lambda v: bool(len(v)), na_action="ignore")
        return arg[mask].explode()

    @classmethod
    def visit(
        cls, op: ops.ElementWiseVectorizedUDF, func, func_args, input_type, return_type
    ):
        """Execute an elementwise UDF."""

        def mapper(df):
            cols = [df[col] for col in df]
            return func(*cols)

        df, _ = cls.asframe(func_args)
        result = df.map_partitions(mapper)
        if op.dtype.is_struct():
            result = result.apply(lambda row: row.to_dict(), axis=1)
        return result

    ############################# Reductions ##################################

    @classmethod
    def visit(cls, op: ops.ArgMin | ops.ArgMax, arg, key, where):
        # TODO(kszucs): raise a warning about triggering compute()?
        if isinstance(op, ops.ArgMin):
            func = lambda x: x.idxmin()
        else:
            func = lambda x: x.idxmax()

        if where is None:

            def agg(df):
                indices = func(df[key.name])
                if isinstance(indices, (dd.Series, dd.core.Scalar)):
                    # to support both aggregating within a group and globally
                    indices = indices.compute()
                return df[arg.name].loc[indices]
        else:

            def agg(df):
                mask = df[where.name]
                filtered = df[mask]
                indices = func(filtered[key.name])
                if isinstance(indices, (dd.Series, dd.core.Scalar)):
                    # to support both aggregating within a group and globally
                    indices = indices.compute()
                return filtered[arg.name].loc[indices]

        return agg

    @classmethod
    def visit(cls, op: ops.Correlation, left, right, where, how):
        if where is None:

            def agg(df):
                return df[left.name].corr(df[right.name])
        else:

            def agg(df):
                mask = df[where.name]
                lhs = df[left.name][mask].compute()
                rhs = df[right.name][mask].compute()
                return lhs.corr(rhs)

        return agg

    @classmethod
    def visit(cls, op: ops.Covariance, left, right, where, how):
        # TODO(kszucs): raise a warning about triggering compute()?
        ddof = {"pop": 0, "sample": 1}[how]
        if where is None:

            def agg(df):
                lhs = df[left.name].compute()
                rhs = df[right.name].compute()
                return lhs.cov(rhs, ddof=ddof)
        else:

            def agg(df):
                mask = df[where.name]
                lhs = df[left.name][mask].compute()
                rhs = df[right.name][mask].compute()
                return lhs.cov(rhs, ddof=ddof)

        return agg

    @classmethod
    def visit(
        cls, op: ops.ReductionVectorizedUDF, func, func_args, input_type, return_type
    ):
        def agg(df):
            # if df is a dask dataframe then we collect it to a pandas dataframe
            # because the user-defined function expects a pandas dataframe
            if isinstance(df, dd.DataFrame):
                df = df.compute()
            args = [df[col.name] for col in func_args]
            return func(*args)

        return agg

    @classmethod
    def visit(
        cls, op: ops.AnalyticVectorizedUDF, func, func_args, input_type, return_type
    ):
        def agg(df, order_keys):
            # if df is a dask dataframe then we collect it to a pandas dataframe
            # because the user-defined function expects a pandas dataframe
            if isinstance(df, dd.DataFrame):
                df = df.compute()
            args = [df[col.name] for col in func_args]
            res = func(*args)
            if isinstance(res, pd.DataFrame):
                # it is important otherwise it is going to fill up the memory
                res = res.apply(lambda row: row.to_dict(), axis=1)
            return res

        return agg

    ############################ Window functions #############################

    @classmethod
    def visit(cls, op: ops.WindowFrame, table, start, end, **kwargs):
        table = table.compute()
        if isinstance(start, dd.Series):
            start = start.compute()
        if isinstance(end, dd.Series):
            end = end.compute()
        return super().visit(op, table=table, start=start, end=end, **kwargs)

    @classmethod
    def visit(cls, op: ops.WindowFunction, func, frame):
        result = super().visit(op, func=func, frame=frame)
        return cls.asseries(result)

    ############################ Relational ###################################

    @classmethod
    def visit(cls, op: ops.DatabaseTable, name, schema, source, namespace):
        try:
            return source.dictionary[name]
        except KeyError:
            raise UnboundExpressionError(
                f"{name} is not a table in the {source.name!r} backend, you "
                "probably tried to execute an expression without a data source"
            )

    @classmethod
    def visit(cls, op: ops.InMemoryTable, name, schema, data):
        df = data.to_frame().reset_index(drop=True)
        return dd.from_pandas(df, npartitions=1)

    @classmethod
    def visit(cls, op: ops.DummyTable, values):
        df, _ = cls.asframe(values)
        return df

    @classmethod
    def visit(cls, op: PandasLimit, parent, n, offset):
        n = n.compute().iat[0, 0]
        offset = offset.compute().iat[0, 0]

        name = gen_name("limit")
        df = add_globally_consecutive_column(parent, name, set_as_index=False)
        if n is None:
            df = df[df[name] >= offset]
        else:
            df = df[df[name].between(offset, offset + n - 1)]

        return df.drop(columns=[name])

    @classmethod
    def visit(cls, op: PandasResetIndex, parent):
        return add_globally_consecutive_column(parent)

    @classmethod
    def visit(cls, op: PandasJoin, **kwargs):
        df = super().visit(op, **kwargs)
        return add_globally_consecutive_column(df)

    @classmethod
    def visit(cls, op: ops.Project, parent, values):
        df, all_scalars = cls.asframe(values)
        if all_scalars and len(parent) != len(df):
            df = dd.concat([df] * len(parent))
        return df

    @classmethod
    def visit(cls, op: ops.Filter, parent, predicates):
        if predicates:
            pred = reduce(operator.and_, predicates)
            parent = parent.loc[pred].reset_index(drop=True)
        return parent

    @classmethod
    def visit(cls, op: ops.Sort, parent, keys):
        # 1. add sort key columns to the dataframe if they are not already present
        # 2. sort the dataframe using those columns
        # 3. drop the sort key columns
        ascending = [key.ascending for key in op.keys]
        newcols = {gen_name("sort_key"): col for col in keys}
        names = list(newcols.keys())
        df = parent.assign(**newcols)
        df = df.sort_values(by=names, ascending=ascending)
        return df.drop(names, axis=1)

    @classmethod
    def visit(cls, op: PandasAggregate, parent, groups, metrics):
        if not groups:
            results = {k: v(parent) for k, v in metrics.items()}
            combined, _ = cls.asframe(results)
            return combined

        parent = parent.groupby([col.name for col in groups.values()])

        measures = {}
        for name, metric in metrics.items():
            meta = pd.Series(
                name=name,
                dtype=PandasType.from_ibis(op.metrics[name].dtype),
                index=pd.MultiIndex(
                    levels=[[] for _ in groups],
                    codes=[[] for _ in groups],
                    names=list(groups.keys()),
                ),
            )
            measures[name] = parent.apply(metric, meta=meta)

        result = cls.concat(measures, axis=1).reset_index()
        renames = {v.name: k for k, v in op.groups.items()}
        return result.rename(columns=renames)

    @classmethod
    def visit(cls, op: ops.InValues, value, options):
        if isinstance(value, dd.Series):
            return value.isin(options)
        else:
            return value in options

    @classmethod
    def visit(cls, op: ops.InSubquery, rel, needle):
        first_column = rel.compute().iloc[:, 0]
        if isinstance(needle, dd.Series):
            return needle.isin(first_column)
        else:
            return needle in first_column

    @classmethod
    def visit(cls, op: PandasScalarSubquery, rel):
        # TODO(kszucs): raise a warning about triggering compute()?
        # could the compute be avoided here?
        return rel.compute().iat[0, 0]

    @classmethod
    def compile(cls, node, backend, params):
        def fn(node, _, **kwargs):
            return cls.visit(node, **kwargs)

        node = node.to_expr().as_table().op()
        node = plan(node, backend=backend, params=params)
        return node.map_clear(fn)

    @classmethod
    def execute(cls, node, backend, params):
        original = node
        node = node.to_expr().as_table().op()
        result = cls.compile(node, backend=backend, params=params)

        # should happen when the result is empty
        if isinstance(result, pd.DataFrame):
            assert result.empty
        else:
            assert isinstance(result, dd.DataFrame)
            result = result.compute()

        result = PandasData.convert_table(result, node.schema)
        if isinstance(original, ops.Value):
            if original.shape.is_scalar():
                return result.iloc[0, 0]
            elif original.shape.is_columnar():
                return result.iloc[:, 0]
            else:
                raise TypeError(f"Unexpected shape: {original.shape}")
        else:
            return result



from __future__ import annotations

import pytest

import ibis
from ibis.common.exceptions import RelationError


class A:
    a: int
    b: str
    c: float


# identical to A
class B:
    a: int
    b: str
    c: float


# same as A with different order
class C:
    c: float
    b: str
    a: int


class D:
    a: str
    b: str
    c: str


a = ibis.table(A)
b = ibis.table(B)
c = ibis.table(C)
d = ibis.table(D)


@pytest.mark.parametrize("method", ["union", "intersect", "difference"])
def test_operation_requires_equal_schemas(method):
    with pytest.raises(RelationError):
        getattr(a, method)(d)


@pytest.mark.parametrize("method", ["union", "intersect", "difference"])
def test_operation_supports_schemas_with_different_field_order(method):
    u1 = getattr(a, method)(b)
    u2 = getattr(a, method)(c)

    assert u1.schema() == a.schema()

    u1 = u1.op().parent
    assert u1.left == a.op()
    assert u1.right == b.op()

    # a selection is added to ensure that the field order of the right table
    # matches the field order of the left table
    u2 = u2.op().parent
    assert u2.schema == a.schema()
    assert u2.left == a.op()

    reprojected = c.select(["a", "b", "c"])
    assert u2.right == reprojected.op()



from __future__ import annotations

import pytest
from pytest import param

import ibis
import ibis.common.exceptions as exc
from ibis import _
from ibis.backends.conftest import _get_backends_to_test

sg = pytest.importorskip("sqlglot")

simple_literal = param(ibis.literal(1), id="simple_literal")
array_literal = param(
    ibis.array([1]),
    marks=[
        pytest.mark.never(
            ["mysql", "mssql", "oracle", "impala", "sqlite"],
            raises=(exc.OperationNotDefinedError, exc.UnsupportedBackendType),
            reason="arrays not supported in the backend",
        ),
    ],
    id="array_literal",
)
no_structs = pytest.mark.never(
    ["impala", "mysql", "sqlite", "mssql", "exasol"],
    raises=(NotImplementedError, exc.UnsupportedBackendType),
    reason="structs not supported in the backend",
)
no_struct_literals = pytest.mark.notimpl(
    ["mssql"], reason="struct literals are not yet implemented"
)
not_sql = pytest.mark.never(
    ["pandas", "dask"],
    raises=(exc.IbisError, NotImplementedError, ValueError),
    reason="Not a SQL backend",
)
no_sql_extraction = pytest.mark.notimpl(
    ["polars"], reason="Not clear how to extract SQL from the backend"
)


@pytest.mark.parametrize(
    "expr",
    [
        simple_literal,
        array_literal,
        param(
            ibis.struct(dict(a=1)),
            marks=[no_structs, no_struct_literals],
            id="struct_literal",
        ),
    ],
)
@not_sql
@no_sql_extraction
def test_literal(backend, expr):
    assert ibis.to_sql(expr, dialect=backend.name())


@pytest.mark.never(["pandas", "dask", "polars"], reason="not SQL")
def test_group_by_has_index(backend, snapshot):
    countries = ibis.table(
        dict(continent="string", population="int64"), name="countries"
    )
    expr = countries.group_by(
        cont=(
            _.continent.case()
            .when("NA", "North America")
            .when("SA", "South America")
            .when("EU", "Europe")
            .when("AF", "Africa")
            .when("AS", "Asia")
            .when("OC", "Oceania")
            .when("AN", "Antarctica")
            .else_("Unknown continent")
            .end()
        )
    ).agg(total_pop=_.population.sum())
    sql = str(ibis.to_sql(expr, dialect=backend.name()))
    snapshot.assert_match(sql, "out.sql")


@pytest.mark.never(["pandas", "dask", "polars"], reason="not SQL")
def test_cte_refs_in_topo_order(backend, snapshot):
    mr0 = ibis.table(schema=ibis.schema(dict(key="int")), name="leaf")

    mr1 = mr0.filter(ibis.literal(True))

    mr2 = mr1.join(mr1[["key"]], ["key"])
    mr3 = mr2.join(mr2, ["key"])

    sql = str(ibis.to_sql(mr3, dialect=backend.name()))
    snapshot.assert_match(sql, "out.sql")


@pytest.mark.never(["pandas", "dask", "polars"], reason="not SQL")
def test_isin_bug(con, snapshot):
    t = ibis.table(dict(x="int"), name="t")
    good = t[t.x > 2].x
    expr = t.x.isin(good)
    snapshot.assert_match(str(ibis.to_sql(expr, dialect=con.name)), "out.sql")


@pytest.mark.never(
    ["pandas", "dask", "polars"],
    reason="not SQL",
    raises=NotImplementedError,
)
@pytest.mark.notyet(
    ["datafusion", "exasol"],
    reason="no unnest support",
    raises=exc.OperationNotDefinedError,
)
@pytest.mark.notyet(
    ["sqlite", "mysql", "druid", "impala", "mssql"], reason="no unnest support upstream"
)
@pytest.mark.notimpl(
    ["oracle", "flink"],
    reason="unnest not yet implemented",
    raises=exc.OperationNotDefinedError,
)
@pytest.mark.parametrize("backend_name", _get_backends_to_test())
def test_union_aliasing(backend_name, snapshot):
    if backend_name == "snowflake":
        pytest.skip(
            "pivot requires random names to make snowflake work properly and cannot be snapshotted"
        )
    from ibis import selectors as s

    t = ibis.table(
        {
            "field_of_study": "string",
            "1970-71": "int64",
            "1975-76": "int64",
            "1980-81": "int64",
            "1985-86": "int64",
            "1990-91": "int64",
            "1995-96": "int64",
            "2000-01": "int64",
            "2005-06": "int64",
            "2010-11": "int64",
            "2011-12": "int64",
            "2012-13": "int64",
            "2013-14": "int64",
            "2014-15": "int64",
            "2015-16": "int64",
            "2016-17": "int64",
            "2017-18": "int64",
            "2018-19": "int64",
            "2019-20": "int64",
        },
        name="humanities",
    )

    years_window = ibis.window(order_by=_.years, group_by=_.field_of_study)
    diff_agg = (
        t.pivot_longer(s.matches(r"\d{4}-\d{2}"), names_to="years", values_to="degrees")
        .mutate(
            earliest_degrees=_.degrees.first().over(years_window),
            latest_degrees=_.degrees.last().over(years_window),
        )
        .mutate(diff=_.latest_degrees - _.earliest_degrees)
        .group_by(_.field_of_study)
        .agg(diff=_.diff.arbitrary())
    )

    top_ten = diff_agg.order_by(_.diff.desc()).limit(10)
    bottom_ten = diff_agg.filter(_.diff < 0).order_by(_.diff).limit(10)

    result = top_ten.union(bottom_ten)

    snapshot.assert_match(str(ibis.to_sql(result, dialect=backend_name)), "out.sql")


def test_union_generates_predictable_aliases(con):
    t = ibis.memtable(
        data=[{"island": "Torgerson", "body_mass_g": 3750, "sex": "male"}]
    )
    sub1 = t.inner_join(t.view(), "island").mutate(island_right=lambda t: t.island)
    sub2 = t.inner_join(t.view(), "sex").mutate(sex_right=lambda t: t.sex)
    expr = ibis.union(sub1, sub2)
    df = con.execute(expr)
    assert len(df) == 2



from __future__ import annotations

import pytest
from datafusion import (
    SessionContext,
)

import ibis
from ibis.backends.conftest import TEST_TABLES


@pytest.fixture
def name_to_path(data_dir):
    return {
        table_name: data_dir / "parquet" / f"{table_name}.parquet"
        for table_name in TEST_TABLES
    }


def test_none_config():
    config = None
    conn = ibis.datafusion.connect(config)
    assert conn.list_tables() == []


def test_str_config(name_to_path):
    config = {name: str(path) for name, path in name_to_path.items()}
    conn = ibis.datafusion.connect(config)
    assert sorted(conn.list_tables()) == sorted(name_to_path)


def test_path_config(name_to_path):
    config = name_to_path
    conn = ibis.datafusion.connect(config)
    assert sorted(conn.list_tables()) == sorted(name_to_path)


def test_context_config(name_to_path):
    ctx = SessionContext()
    for name, path in name_to_path.items():
        ctx.register_parquet(name, str(path))
    conn = ibis.datafusion.connect(ctx)
    assert sorted(conn.list_tables()) == sorted(name_to_path)



from __future__ import annotations

import abc
import functools
import inspect
import re
from collections import defaultdict

from ibis.common.typing import (
    Union,
    UnionType,
    evaluate_annotations,
    get_args,
    get_origin,
)
from ibis.util import import_object, unalias_package


def normalize(r: str | re.Pattern):
    """Normalize a expression by wrapping it with `'^'` and `'$'`.

    Parameters
    ----------
    r
        The pattern to normalize.

    Returns
    -------
    Pattern
        The compiled regex.

    """
    r = getattr(r, "pattern", r)
    return re.compile("^" + r.lstrip("^").rstrip("$") + "$")


class SingleDispatch:
    def __init__(self, func, typ=None):
        self.lookup = {}
        self.abc_lookup = {}
        self.lazy_lookup = defaultdict(dict)
        self.func = func
        self.add(func, typ)

    def add(self, func, typ=None):
        if typ is None:
            annots = getattr(func, "__annotations__", {})
            typehints = evaluate_annotations(annots, func.__module__, best_effort=True)
            if typehints:
                typ, *_ = typehints.values()
                if get_origin(typ) in (Union, UnionType):
                    for t in get_args(typ):
                        self.add(func, t)
                else:
                    self.add(func, typ)
            else:
                self.add(func, object)
        elif isinstance(typ, tuple):
            for t in typ:
                self.add(func, t)
        elif isinstance(typ, abc.ABCMeta):
            if typ in self.abc_lookup:
                raise TypeError(f"{typ} is already registered")
            self.abc_lookup[typ] = func
        elif isinstance(typ, str):
            package, rest = typ.split(".", 1)
            package = unalias_package(package)
            typ = f"{package}.{rest}"
            if typ in self.lazy_lookup[package]:
                raise TypeError(f"{typ} is already registered")
            self.lazy_lookup[package][typ] = func
        else:
            if typ in self.lookup:
                raise TypeError(f"{typ} is already registered")
            self.lookup[typ] = func
        return func

    def register(self, typ, func=None):
        """Register a new implementation for arguments of type `cls`."""

        def inner(func):
            self.add(func, typ)
            return func

        return inner if func is None else inner(func)

    def dispatch(self, typ):
        """Return the implementation for the given `cls`."""
        for klass in typ.__mro__:
            # 1. Check for a concrete implementation
            try:
                impl = self.lookup[klass]
            except KeyError:
                pass
            else:
                if typ is not klass:
                    # Cache implementation
                    self.lookup[typ] = impl
                return impl
            # 2. Check lazy implementations
            package = klass.__module__.split(".", 1)[0]
            if lazy := self.lazy_lookup.get(package):
                # Import all lazy implementations first before registering
                # (which should never fail), to ensure an error anywhere
                # doesn't result in a half-registered state.
                new = {import_object(name): func for name, func in lazy.items()}
                self.lookup.update(new)
                # drop lazy implementations, idempotent for thread safety
                self.lazy_lookup.pop(package, None)
                return self.dispatch(typ)
            # 3. Check for abcs
            for abc_class, impl in self.abc_lookup.items():
                if issubclass(typ, abc_class):
                    self.lookup[typ] = impl
                    return impl
        raise TypeError(f"Could not find implementation for {typ}")

    def __call__(self, arg, *args, **kwargs):
        impl = self.dispatch(type(arg))
        return impl(arg, *args, **kwargs)

    def __get__(self, obj, cls=None):
        def _method(*args, **kwargs):
            method = self.dispatch(type(args[0]))
            method = method.__get__(obj, cls)
            return method(*args, **kwargs)

        functools.update_wrapper(_method, self.func)
        return _method


def lazy_singledispatch(func):
    """A `singledispatch` implementation that supports lazily registering implementations."""

    dispatcher = SingleDispatch(func, object)

    @functools.wraps(func)
    def call(arg, *args, **kwargs):
        impl = dispatcher.dispatch(type(arg))
        return impl(arg, *args, **kwargs)

    call.dispatch = dispatcher.dispatch
    call.register = dispatcher.register
    return call


class _MultiDict(dict):
    """A dictionary that allows multiple values for a single key."""

    def __setitem__(self, key, value):
        if key in self:
            self[key].append(value)
        else:
            super().__setitem__(key, [value])


class DispatchedMeta(type):
    """Metaclass that allows multiple implementations of a method to be defined."""

    def __new__(cls, name, bases, dct):
        namespace = {}
        for key, value in dct.items():
            if len(value) == 1:
                # there is just a single attribute so pick that
                namespace[key] = value[0]
            elif all(inspect.isfunction(v) for v in value):
                # multiple functions are defined with the same name, so create
                # a dispatcher function
                first, *rest = value
                func = SingleDispatch(first)
                for impl in rest:
                    func.add(impl)
                namespace[key] = func
            elif all(isinstance(v, classmethod) for v in value):
                first, *rest = value
                func = SingleDispatch(first.__func__)
                for impl in rest:
                    func.add(impl.__func__)
                namespace[key] = classmethod(func)
            elif all(isinstance(v, staticmethod) for v in value):
                first, *rest = value
                func = SingleDispatch(first.__func__)
                for impl in rest:
                    func.add(impl.__func__)
                namespace[key] = staticmethod(func)
            else:
                raise TypeError(f"Multiple attributes are defined with name {key}")

        return type.__new__(cls, name, bases, namespace)

    @classmethod
    def __prepare__(cls, name, bases):
        return _MultiDict()


class Dispatched(metaclass=DispatchedMeta):
    """Base class supporting multiple implementations of a method.

    Methods with the same name can be defined multiple times. The first method
    defined is the default implementation, and subsequent methods are registered
    as implementations for specific types of the first argument.

    The constructed methods are equivalent as if they were defined with
    `functools.singledispatchmethod` but without the need to use the decorator
    syntax. The recommended application of this class is to implement visitor
    patterns.

    Besides ordinary methods, classmethods and staticmethods are also supported.
    The implementation can be extended to overload multiple arguments by using
    `multimethod` instead of `singledispatchmethod` as the dispatcher.
    """



from __future__ import annotations

import itertools
from functools import partial

import sqlglot as sg
import sqlglot.expressions as sge
from public import public

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis import util
from ibis.backends.base.sqlglot.compiler import NULL, C, FuncGen, SQLGlotCompiler
from ibis.backends.base.sqlglot.datatypes import SnowflakeType
from ibis.backends.base.sqlglot.dialects import Snowflake
from ibis.backends.base.sqlglot.rewrites import (
    exclude_unsupported_window_frame_from_ops,
    exclude_unsupported_window_frame_from_row_number,
    replace_log2,
    replace_log10,
    rewrite_empty_order_by_window,
    rewrite_first_to_first_value,
    rewrite_last_to_last_value,
)


class SnowflakeFuncGen(FuncGen):
    udf = FuncGen(namespace="ibis_udfs.public")


@public
class SnowflakeCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = Snowflake
    type_mapper = SnowflakeType
    no_limit_value = NULL
    rewrites = (
        exclude_unsupported_window_frame_from_row_number,
        exclude_unsupported_window_frame_from_ops,
        rewrite_first_to_first_value,
        rewrite_last_to_last_value,
        rewrite_empty_order_by_window,
        replace_log2,
        replace_log10,
        *SQLGlotCompiler.rewrites,
    )

    UNSUPPORTED_OPERATIONS = frozenset(
        (
            ops.ArrayMap,
            ops.ArrayFilter,
            ops.RowID,
            ops.MultiQuantile,
            ops.IntervalFromInteger,
            ops.IntervalAdd,
            ops.TimestampDiff,
        )
    )

    SIMPLE_OPS = {
        ops.Any: "max",
        ops.All: "min",
        ops.ArrayDistinct: "array_distinct",
        ops.ArrayFlatten: "array_flatten",
        ops.ArrayIndex: "get",
        ops.ArrayIntersect: "array_intersection",
        ops.ArrayRemove: "array_remove",
        ops.BitAnd: "bitand_agg",
        ops.BitOr: "bitor_agg",
        ops.BitXor: "bitxor_agg",
        ops.BitwiseAnd: "bitand",
        ops.BitwiseLeftShift: "bitshiftleft",
        ops.BitwiseNot: "bitnot",
        ops.BitwiseOr: "bitor",
        ops.BitwiseRightShift: "bitshiftright",
        ops.BitwiseXor: "bitxor",
        ops.EndsWith: "endswith",
        ops.Hash: "hash",
        ops.Median: "median",
        ops.Mode: "mode",
        ops.StringToTimestamp: "to_timestamp_tz",
        ops.TimeFromHMS: "time_from_parts",
        ops.TimestampFromYMDHMS: "timestamp_from_parts",
    }

    def __init__(self):
        super().__init__()
        self.f = SnowflakeFuncGen()

    def _aggregate(self, funcname: str, *args, where):
        if where is not None:
            args = [self.if_(where, arg, NULL) for arg in args]

        func = self.f[funcname]
        return func(*args)

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    def visit_Literal(self, op, *, value, dtype):
        if value is None:
            return super().visit_Literal(op, value=value, dtype=dtype)
        elif dtype.is_string():
            # sqlglot doesn't escape backslashes in strings
            return sge.convert(value.replace("\\", "\\\\"))
        elif dtype.is_timestamp():
            args = (
                value.year,
                value.month,
                value.day,
                value.hour,
                value.minute,
                value.second,
                value.microsecond * 1_000,
            )
            if value.tzinfo is not None:
                return self.f.timestamp_tz_from_parts(*args, dtype.timezone)
            else:
                # workaround sqlglot not supporting more than 6 arguments by
                # using an anonymous function
                return self.f.anon.timestamp_from_parts(*args)
        elif dtype.is_time():
            nanos = value.microsecond * 1_000
            return self.f.time_from_parts(value.hour, value.minute, value.second, nanos)
        elif dtype.is_map():
            key_type = dtype.key_type
            value_type = dtype.value_type

            pairs = []

            for k, v in value.items():
                pairs.append(
                    self.visit_Literal(
                        ops.Literal(k, key_type), value=k, dtype=key_type
                    )
                )
                pairs.append(
                    self.visit_Literal(
                        ops.Literal(v, value_type), value=v, dtype=value_type
                    )
                )

            return self.f.object_construct_keep_null(*pairs)
        elif dtype.is_struct():
            pairs = []
            for k, v in value.items():
                pairs.append(k)
                pairs.append(
                    self.visit_Literal(
                        ops.Literal(v, dtype[k]), value=v, dtype=dtype[k]
                    )
                )
            return self.f.object_construct_keep_null(*pairs)
        elif dtype.is_uuid():
            return sge.convert(str(value))
        elif dtype.is_binary():
            return sge.HexString(this=value.hex())
        return super().visit_Literal(op, value=value, dtype=dtype)

    def visit_Cast(self, op, *, arg, to):
        if to.is_struct() or to.is_map():
            return self.if_(self.f.is_object(arg), arg, NULL)
        elif to.is_array():
            return self.if_(self.f.is_array(arg), arg, NULL)
        return self.cast(arg, to)

    def visit_ToJSONMap(self, op, *, arg):
        return self.if_(self.f.is_object(arg), arg, NULL)

    def visit_ToJSONArray(self, op, *, arg):
        return self.if_(self.f.is_array(arg), arg, NULL)

    def visit_IsNan(self, op, *, arg):
        return arg.eq(self.NAN)

    def visit_IsInf(self, op, *, arg):
        return arg.isin(self.POS_INF, self.NEG_INF)

    def visit_JSONGetItem(self, op, *, arg, index):
        return self.f.get(arg, index)

    def visit_StringFind(self, op, *, arg, substr, start, end):
        args = [substr, arg]
        if start is not None:
            args.append(start + 1)
        return self.f.position(*args)

    def visit_RegexSplit(self, op, *, arg, pattern):
        return self.f.udf.regexp_split(arg, pattern)

    def visit_Map(self, op, *, keys, values):
        return self.if_(
            sg.and_(self.f.is_array(keys), self.f.is_array(values)),
            self.f.udf.object_from_arrays(keys, values),
            NULL,
        )

    def visit_MapKeys(self, op, *, arg):
        return self.if_(self.f.is_object(arg), self.f.object_keys(arg), NULL)

    def visit_MapValues(self, op, *, arg):
        return self.if_(self.f.is_object(arg), self.f.udf.object_values(arg), NULL)

    def visit_MapGet(self, op, *, arg, key, default):
        dtype = op.dtype
        expr = self.f.coalesce(self.f.get(arg, key), self.f.to_variant(default))
        if dtype.is_json() or dtype.is_null():
            return expr
        return self.cast(expr, dtype)

    def visit_MapContains(self, op, *, arg, key):
        return self.f.array_contains(
            self.if_(self.f.is_object(arg), self.f.object_keys(arg), NULL),
            self.f.to_variant(key),
        )

    def visit_MapMerge(self, op, *, left, right):
        return self.if_(
            sg.and_(self.f.is_object(left), self.f.is_object(right)),
            self.f.udf.object_merge(left, right),
            NULL,
        )

    def visit_MapLength(self, op, *, arg):
        return self.if_(
            self.f.is_object(arg), self.f.array_size(self.f.object_keys(arg)), NULL
        )

    def visit_Log(self, op, *, arg, base):
        return self.f.log(base, arg, dialect=self.dialect)

    def visit_RandomScalar(self, op):
        return self.f.uniform(
            self.f.to_double(0.0), self.f.to_double(1.0), self.f.random()
        )

    def visit_ApproxMedian(self, op, *, arg, where):
        return self.agg.approx_percentile(arg, 0.5, where=where)

    def visit_TimeDelta(self, op, *, part, left, right):
        return self.f.timediff(part, right, left, dialect=self.dialect)

    def visit_DateDelta(self, op, *, part, left, right):
        return self.f.datediff(part, right, left, dialect=self.dialect)

    def visit_TimestampDelta(self, op, *, part, left, right):
        return self.f.timestampdiff(part, right, left, dialect=self.dialect)

    def visit_TimestampDateAdd(self, op, *, left, right):
        if not isinstance(op.right, ops.Literal):
            raise com.OperationNotDefinedError(
                f"right side of {type(op).__name__} operation must be an interval literal"
            )
        return sg.exp.Add(this=left, expression=right)

    visit_DateAdd = visit_TimestampAdd = visit_TimestampDateAdd

    def visit_IntegerRange(self, op, *, start, stop, step):
        return self.if_(
            step.neq(0), self.f.array_generate_range(start, stop, step), self.f.array()
        )

    def visit_StructColumn(self, op, *, names, values):
        return self.f.object_construct_keep_null(
            *itertools.chain.from_iterable(zip(names, values))
        )

    def visit_StructField(self, op, *, arg, field):
        return self.cast(self.f.get(arg, field), op.dtype)

    def visit_RegexSearch(self, op, *, arg, pattern):
        return sge.RegexpLike(
            this=arg,
            expression=self.f.concat(".*", pattern, ".*"),
            flag=sge.convert("cs"),
        )

    def visit_RegexReplace(self, op, *, arg, pattern, replacement):
        return sge.RegexpReplace(this=arg, expression=pattern, replacement=replacement)

    def visit_TypeOf(self, op, *, arg):
        return self.f.typeof(self.f.to_variant(arg))

    def visit_ArrayRepeat(self, op, *, arg, times):
        return self.f.udf.array_repeat(arg, times)

    def visit_ArrayUnion(self, op, *, left, right):
        return self.f.array_distinct(self.f.array_cat(left, right))

    def visit_ArrayContains(self, op, *, arg, other):
        return self.f.array_contains(arg, self.f.to_variant(other))

    def visit_ArrayCollect(self, op, *, arg, where):
        return self.agg.array_agg(
            self.f.ifnull(arg, self.f.parse_json("null")), where=where
        )

    def visit_ArrayConcat(self, op, *, arg):
        # array_cat only accepts two arguments
        return self.f.array_flatten(self.f.array(*arg))

    def visit_ArrayPosition(self, op, *, arg, other):
        # snowflake is zero-based here, so we don't need to subtract 1 from the
        # result
        return self.f.coalesce(
            self.f.array_position(self.f.to_variant(other), arg) + 1, 0
        )

    def visit_RegexExtract(self, op, *, arg, pattern, index):
        # https://docs.snowflake.com/en/sql-reference/functions/regexp_substr
        return sge.RegexpExtract(
            this=arg,
            expression=pattern,
            position=sge.convert(1),
            group=index,
            parameters=sge.convert("ce"),
        )

    def visit_ArrayZip(self, op, *, arg):
        return self.f.udf.array_zip(self.f.array(*arg))

    def visit_DayOfWeekName(self, op, *, arg):
        return sge.Case(
            this=self.f.dayname(arg),
            ifs=[
                self.if_("Sun", "Sunday"),
                self.if_("Mon", "Monday"),
                self.if_("Tue", "Tuesday"),
                self.if_("Wed", "Wednesday"),
                self.if_("Thu", "Thursday"),
                self.if_("Fri", "Friday"),
                self.if_("Sat", "Saturday"),
            ],
            default=NULL,
        )

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        timestamp_units_to_scale = {"s": 0, "ms": 3, "us": 6, "ns": 9}
        return self.f.to_timestamp(arg, timestamp_units_to_scale[unit.short])

    def visit_First(self, op, *, arg, where):
        return self.f.get(self.agg.array_agg(arg, where=where), 0)

    def visit_Last(self, op, *, arg, where):
        expr = self.agg.array_agg(arg, where=where)
        return self.f.get(expr, self.f.array_size(expr) - 1)

    def visit_GroupConcat(self, op, *, arg, where, sep):
        if where is None:
            return self.f.listagg(arg, sep)

        return self.if_(
            self.f.count_if(where) > 0,
            self.f.listagg(self.if_(where, arg, NULL), sep),
            NULL,
        )

    def visit_TimestampBucket(self, op, *, arg, interval, offset):
        if offset is not None:
            raise com.UnsupportedOperationError(
                "`offset` is not supported in the Snowflake backend for timestamp bucketing"
            )

        interval = op.interval
        if not isinstance(interval, ops.Literal):
            raise com.UnsupportedOperationError(
                f"Interval must be a literal for the Snowflake backend, got {type(interval)}"
            )

        return self.f.time_slice(arg, interval.value, interval.dtype.unit.name)

    def visit_Arbitrary(self, op, *, arg, how, where):
        if how == "first":
            return self.f.get(self.agg.array_agg(arg, where=where), 0)
        elif how == "last":
            expr = self.agg.array_agg(arg, where=where)
            return self.f.get(expr, self.f.array_size(expr) - 1)
        else:
            raise com.UnsupportedOperationError("how must be 'first' or 'last'")

    def visit_ArraySlice(self, op, *, arg, start, stop):
        if start is None:
            start = 0

        if stop is None:
            stop = self.f.array_size(arg)
        return self.f.array_slice(arg, start, stop)

    def visit_ExtractEpochSeconds(self, op, *, arg):
        return self.f.extract("epoch", arg)

    def visit_ExtractMicrosecond(self, op, *, arg):
        return self.f.extract("epoch_microsecond", arg) % 1_000_000

    def visit_ExtractMillisecond(self, op, *, arg):
        return self.f.extract("epoch_millisecond", arg) % 1_000

    def visit_ExtractQuery(self, op, *, arg, key):
        parsed_url = self.f.parse_url(arg, 1)
        if key is not None:
            r = self.f.get(self.f.get(parsed_url, "parameters"), key)
        else:
            r = self.f.get(parsed_url, "query")
        return self.f.nullif(self.f.as_varchar(r), "")

    def visit_ExtractProtocol(self, op, *, arg):
        return self.f.nullif(
            self.f.as_varchar(self.f.get(self.f.parse_url(arg, 1), "scheme")), ""
        )

    def visit_ExtractAuthority(self, op, *, arg):
        return self.f.concat_ws(
            ":",
            self.f.as_varchar(self.f.get(self.f.parse_url(arg, 1), "host")),
            self.f.as_varchar(self.f.get(self.f.parse_url(arg, 1), "port")),
        )

    def visit_ExtractFile(self, op, *, arg):
        return self.f.concat_ws(
            "?",
            self.visit_ExtractPath(op, arg=arg),
            self.visit_ExtractQuery(op, arg=arg, key=None),
        )

    def visit_ExtractPath(self, op, *, arg):
        return self.f.concat(
            "/", self.f.as_varchar(self.f.get(self.f.parse_url(arg, 1), "path"))
        )

    def visit_ExtractFragment(self, op, *, arg):
        return self.f.nullif(
            self.f.as_varchar(self.f.get(self.f.parse_url(arg, 1), "fragment")), ""
        )

    def visit_Unnest(self, op, *, arg):
        sep = sge.convert(util.guid())
        split = self.f.split(
            self.f.array_to_string(self.f.nullif(arg, self.f.array()), sep), sep
        )
        expr = self.f.nullif(self.f.explode(split), "")
        return self.cast(expr, op.dtype)

    def visit_Quantile(self, op, *, arg, quantile, where):
        # can't use `self.agg` here because `quantile` must be a constant and
        # the agg method filters using `where` for every argument which turns
        # the constant into an expression
        if where is not None:
            arg = self.if_(where, arg, NULL)
        return self.f.percentile_cont(arg, quantile)

    def visit_CountStar(self, op, *, arg, where):
        if where is None:
            return super().visit_CountStar(op, arg=arg, where=where)
        return self.f.count_if(where)

    def visit_CountDistinct(self, op, *, arg, where):
        if where is not None:
            arg = self.if_(where, arg, NULL)
        return self.f.count(sge.Distinct(expressions=[arg]))

    def visit_CountDistinctStar(self, op, *, arg, where):
        columns = op.arg.schema.names
        quoted = self.quoted
        col = partial(sg.column, quoted=quoted)
        if where is None:
            expressions = list(map(col, columns))
        else:
            # any null columns will cause the entire row not to be counted
            expressions = [self.if_(where, col(name), NULL) for name in columns]
        return self.f.count(sge.Distinct(expressions=expressions))

    def visit_Xor(self, op, *, left, right):
        # boolxor accepts numerics ... and returns a boolean? wtf?
        return self.f.boolxor(self.cast(left, dt.int8), self.cast(right, dt.int8))

    def visit_WindowBoundary(self, op, *, value, preceding):
        if not isinstance(op.value, ops.Literal):
            raise com.OperationNotDefinedError(
                "Expressions in window bounds are not supported by Snowflake"
            )
        return super().visit_WindowBoundary(op, value=value, preceding=preceding)

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "sample":
            raise com.UnsupportedOperationError(
                f"{self.dialect} only implements `pop` correlation coefficient"
            )

        # TODO: rewrite rule?
        if (left_type := op.left.dtype).is_boolean():
            left = self.cast(left, dt.Int32(nullable=left_type.nullable))

        if (right_type := op.right.dtype).is_boolean():
            right = self.cast(right, dt.Int32(nullable=right_type.nullable))

        return self.agg.corr(left, right, where=where)

    def visit_TimestampRange(self, op, *, start, stop, step):
        raw_step = op.step

        if not isinstance(raw_step, ops.Literal):
            raise com.UnsupportedOperationError("`step` argument must be a literal")

        unit = raw_step.dtype.unit.name.lower()
        step = raw_step.value

        value_type = op.dtype.value_type

        if step == 0:
            return self.f.array()

        return (
            sg.select(
                self.f.array_agg(
                    self.f.replace(
                        # conversion to varchar is necessary to control
                        # the timestamp format
                        #
                        # otherwise, since timestamps in arrays become strings
                        # anyway due to lack of parameterized type support in
                        # Snowflake the format depends on a session parameter
                        self.f.to_varchar(
                            self.f.dateadd(unit, C.value, start, dialect=self.dialect),
                            'YYYY-MM-DD"T"HH24:MI:SS.FF6'
                            + (value_type.timezone is not None) * "TZH:TZM",
                        ),
                        # timezones are always hour:minute offsets from UTC, not
                        # named, so replacing "Z" shouldn't be an issue
                        "Z",
                        "+00:00",
                    ),
                )
            )
            .from_(
                sge.Table(
                    this=sge.Unnest(
                        expressions=[
                            self.f.array_generate_range(
                                0,
                                self.f.datediff(
                                    unit, start, stop, dialect=self.dialect
                                ),
                                step,
                            )
                        ]
                    )
                )
            )
            .subquery()
        )



from __future__ import annotations

import itertools
from urllib.parse import parse_qs, urlsplit

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow_hotfix  # noqa: F401

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt


def _extract_epoch_seconds(array) -> dt.int32:
    return pc.cast(pc.divide(pc.cast(array, pa.int64()), 1_000_000), pa.int32())


def extract_epoch_seconds_date(array: dt.date) -> dt.int32:
    return _extract_epoch_seconds(array)


def extract_epoch_seconds_timestamp(array: dt.Timestamp(scale=6)) -> dt.int32:
    return _extract_epoch_seconds(array)


def _extract_second(array):
    return pc.cast(pc.second(array), pa.int32())


def extract_second_timestamp(array: dt.Timestamp(scale=9)) -> dt.int32:
    return _extract_second(array)


def extract_second_time(array: dt.time) -> dt.int32:
    return _extract_second(array)


def _extract_millisecond(array) -> dt.int32:
    return pc.cast(pc.millisecond(array), pa.int32())


def extract_millisecond_timestamp(array: dt.Timestamp(scale=9)) -> dt.int32:
    return _extract_millisecond(array)


def extract_millisecond_time(array: dt.time) -> dt.int32:
    return _extract_millisecond(array)


def extract_microsecond(array: dt.Timestamp(scale=9)) -> dt.int32:
    arr = pc.multiply(pc.millisecond(array), 1000)
    return pc.cast(pc.add(pc.microsecond(array), arr), pa.int32())


def _extract_query_arrow(
    arr: pa.StringArray, *, param: str | None = None
) -> pa.StringArray:
    if param is None:

        def _extract_query(url, param):
            return urlsplit(url).query

        params = itertools.repeat(None)
    else:

        def _extract_query(url, param):
            query = urlsplit(url).query
            value = parse_qs(query)[param]
            return value if len(value) > 1 else value[0]

        params = param.to_pylist()

    return pa.array(map(_extract_query, arr.to_pylist(), params))


def extract_query(array: str) -> str:
    return _extract_query_arrow(array)


def extract_query_param(array: str, param: str) -> str:
    return _extract_query_arrow(array, param=param)


def extract_user_info(arr: str) -> str:
    def _extract_user_info(url):
        url_parts = urlsplit(url)
        username = url_parts.username or ""
        password = url_parts.password or ""
        return f"{username}:{password}"

    return pa.array(map(_extract_user_info, arr.to_pylist()))


def extract_url_field(arr: str, field: str) -> str:
    field = field.to_pylist()[0]
    return pa.array(getattr(url, field, "") for url in map(urlsplit, arr.to_pylist()))


def sign(arr: dt.float64) -> dt.float64:
    return pc.sign(arr)


def _extract_minute(array) -> dt.int32:
    return pc.cast(pc.minute(array), pa.int32())


def extract_minute_time(array: dt.time) -> dt.int32:
    return _extract_minute(array)


def extract_minute_timestamp(array: dt.Timestamp(scale=9)) -> dt.int32:
    return _extract_minute(array)


def extract_hour_time(array: dt.time) -> dt.int32:
    return pc.cast(pc.hour(array), pa.int32())


def regex_split(s: str, pattern: str) -> list[str]:
    # TODO: pretty inefficient, but this is a stopgap until we can get an
    # upstream version of this function
    #
    # unique is necessary because when `s` is coming from a column, `pattern`
    # is repeated to match the length of `s`
    patterns = pattern.unique()
    if len(patterns) != 1:
        raise com.IbisError(
            "Only a single scalar pattern is supported for DataFusion re_split"
        )
    pattern = patterns[0].as_py()
    return pc.split_pattern_regex(s, pattern)



from __future__ import annotations

import sqlite3
from datetime import date

import pandas as pd
import pytest
from packaging.version import parse as vparse

import ibis
import ibis.expr.datatypes as dt

# Test with formats 1-7 (with T, Z, and offset modifiers) from:
# https://sqlite.org/lang_datefunc.html#time_values
TIMESTAMPS = [
    "2022-01-02",
    "2022-01-02 03:04",
    "2022-01-02 03:04:05",
    "2022-01-02 03:04:05.678",
    "2022-01-02T03:04",
    "2022-01-02T03:04:05",
    "2022-01-02T03:04:05.678",
    None,
]
TIMESTAMPS_TZ = [
    "2022-01-02 03:04Z",
    "2022-01-02 03:04:05Z",
    "2022-01-02 03:04:05.678Z",
    "2022-01-02 03:04+01:00",
    "2022-01-02 03:04:05+01:00",
    "2022-01-02 03:04:05.678+01:00",
    None,
]


@pytest.fixture(scope="session")
def db(tmp_path_factory):
    path = str(tmp_path_factory.mktemp("databases") / "formats.db")
    con = sqlite3.connect(path)
    con.execute("CREATE TABLE timestamps (ts TIMESTAMP)")
    con.execute("CREATE TABLE timestamps_tz (ts TIMESTAMPTZ)")
    con.execute("CREATE TABLE weird (str_col STRING, date_col ITSADATE)")
    con.execute("CREATE TABLE basic (a INTEGER, b REAL, c BOOLEAN, d BLOB)")
    with con:
        con.executemany("INSERT INTO timestamps VALUES (?)", [(t,) for t in TIMESTAMPS])
        con.executemany(
            "INSERT INTO timestamps_tz VALUES (?)",
            [(t,) for t in TIMESTAMPS_TZ],
        )
        con.executemany(
            "INSERT INTO weird VALUES (?, ?)",
            [
                ("a", "2022-01-01"),
                ("b", "2022-01-02"),
                ("c", "2022-01-03"),
                ("d", "2022-01-04"),
            ],
        )
    con.close()
    return path


@pytest.mark.parametrize(
    "table, data",
    [("timestamps", TIMESTAMPS), ("timestamps_tz", TIMESTAMPS_TZ)],
)
def test_timestamps(db, table, data):
    con = ibis.sqlite.connect(db)
    t = con.table(table)
    assert t.ts.type().is_timestamp()
    res = t.ts.execute()
    # the "mixed" format was added in pandas 2.0.0
    format = "mixed" if vparse(pd.__version__) >= vparse("2.0.0") else None
    stamps = pd.to_datetime(data, format=format, utc=True)
    if t.ts.type().timezone is None:
        # we're casting to timestamp without a timezone, so remove it in the
        # expected output
        stamps = stamps.tz_localize(None)
    sol = pd.Series(stamps)
    assert res.equals(sol)


def test_type_map(db):
    con = ibis.sqlite.connect(db, type_map={"STRING": dt.string, "ITSADATE": "date"})
    t = con.table("weird")
    expected_schema = ibis.schema({"str_col": "string", "date_col": "date"})
    assert t.schema() == expected_schema
    res = t.filter(t.str_col == "a").execute()
    sol = pd.DataFrame(
        {"str_col": ["a"], "date_col": pd.Series([date(2022, 1, 1)], dtype="object")}
    )
    assert res.equals(sol)


def test_read_basic_types(db):
    con = ibis.sqlite.connect(db)
    t = con.table("basic")
    assert t.schema() == ibis.schema(
        {"a": "int64", "b": "float64", "c": "bool", "d": "binary"}
    )



from __future__ import annotations

from .conftest import tpch_test


@tpch_test
def test_tpc_h19(lineitem, part):
    """Discounted Revenue Query (Q19)

    The Discounted Revenue Query reports the gross discounted revenue
    attributed to the sale of selected parts handled in a particular manner.
    This query is an example of code such as might be produced programmatically
    by a data mining tool."""

    QUANTITY1 = 1
    QUANTITY2 = 10
    QUANTITY3 = 20
    BRAND1 = "Brand#12"
    BRAND2 = "Brand#23"
    BRAND3 = "Brand#34"

    q = lineitem.join(part, part.p_partkey == lineitem.l_partkey)

    q1 = (
        (q.p_brand == BRAND1)
        & (q.p_container.isin(("SM CASE", "SM BOX", "SM PACK", "SM PKG")))
        & (q.l_quantity >= QUANTITY1)
        & (q.l_quantity <= QUANTITY1 + 10)
        & (q.p_size.between(1, 5))
        & (q.l_shipmode.isin(("AIR", "AIR REG")))
        & (q.l_shipinstruct == "DELIVER IN PERSON")
    )

    q2 = (
        (q.p_brand == BRAND2)
        & (q.p_container.isin(("MED BAG", "MED BOX", "MED PKG", "MED PACK")))
        & (q.l_quantity >= QUANTITY2)
        & (q.l_quantity <= QUANTITY2 + 10)
        & (q.p_size.between(1, 10))
        & (q.l_shipmode.isin(("AIR", "AIR REG")))
        & (q.l_shipinstruct == "DELIVER IN PERSON")
    )

    q3 = (
        (q.p_brand == BRAND3)
        & (q.p_container.isin(("LG CASE", "LG BOX", "LG PACK", "LG PKG")))
        & (q.l_quantity >= QUANTITY3)
        & (q.l_quantity <= QUANTITY3 + 10)
        & (q.p_size.between(1, 15))
        & (q.l_shipmode.isin(("AIR", "AIR REG")))
        & (q.l_shipinstruct == "DELIVER IN PERSON")
    )

    q = q.filter([q1 | q2 | q3])
    q = q.aggregate(revenue=(q.l_extendedprice * (1 - q.l_discount)).sum())
    return q



from __future__ import annotations

import numpy as np

from ibis.formats.pandas import PandasData


class DuckDBPandasData(PandasData):
    @staticmethod
    def convert_Array(s, dtype, pandas_type):
        return s.replace(np.nan, None)



from __future__ import annotations

import pandas as pd
import pandas.testing as tm
import pytest
from pytest import param

import ibis
from ibis.common.exceptions import IbisTypeError

pyspark = pytest.importorskip("pyspark")


@pytest.fixture
def t(con):
    return con.table("basic_table")


@pytest.fixture
def df(con):
    return con._session.table("basic_table").toPandas()


def test_basic(t):
    result = t.execute()
    expected = pd.DataFrame({"id": range(10), "str_col": "value"})
    tm.assert_frame_equal(result, expected)


def test_projection(t):
    result1 = t.mutate(v=t["id"]).execute()
    expected1 = pd.DataFrame({"id": range(10), "str_col": "value", "v": range(10)})

    result2 = t.mutate(v=t["id"]).mutate(v2=t["id"]).mutate(id=t["id"] * 2).execute()
    expected2 = pd.DataFrame(
        {
            "id": range(0, 20, 2),
            "str_col": "value",
            "v": range(10),
            "v2": range(10),
        }
    )

    tm.assert_frame_equal(result1, expected1)
    tm.assert_frame_equal(result2, expected2)


def test_aggregation_col(t, df):
    result = t["id"].count().execute()
    assert result == len(df)


def test_aggregation(t, df):
    result = t.aggregate(max=t["id"].max()).execute()
    expected = pd.DataFrame({"max": [df.id.max()]})
    tm.assert_frame_equal(result, expected)


def test_group_by(t, df):
    result = t.group_by("id").aggregate(max=t["id"].max()).execute()
    expected = df[["id"]].assign(max=df.groupby("id").id.max())
    tm.assert_frame_equal(result, expected)


def test_window(t, df):
    w = ibis.window()
    result = t.mutate(grouped_demeaned=t["id"] - t["id"].mean().over(w)).execute()
    expected = df.assign(grouped_demeaned=df.id - df.id.mean())

    tm.assert_frame_equal(result, expected)


def test_greatest(t, df):
    result = t.mutate(greatest=ibis.greatest(t.id, t.id + 1)).execute()
    expected = df.assign(greatest=df.id + 1)
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    ("filter_fn", "expected_fn"),
    [
        param(lambda t: t.filter(t.id < 5), lambda df: df[df.id < 5]),
        param(lambda t: t.filter(t.id != 5), lambda df: df[df.id != 5]),
        param(
            lambda t: t.filter([t.id < 5, t.str_col == "na"]),
            lambda df: df[df.id < 5][df.str_col == "na"],
        ),
        param(
            lambda t: t.filter((t.id > 3) & (t.id < 11)),
            lambda df: df[(df.id > 3) & (df.id < 11)],
        ),
        param(
            lambda t: t.filter((t.id == 3) | (t.id == 5)),
            lambda df: df[(df.id == 3) | (df.id == 5)],
        ),
    ],
)
def test_filter(t, df, filter_fn, expected_fn):
    result = filter_fn(t).execute().reset_index(drop=True)
    expected = expected_fn(df).reset_index(drop=True)
    tm.assert_frame_equal(result, expected)


def test_cast(t, df):
    result = t.mutate(id_string=t.id.cast("string")).execute()
    df = df.assign(id_string=df.id.astype(str))
    tm.assert_frame_equal(result, df)


def test_alias_after_select(t, df):
    # Regression test for issue 2136
    table = t[["id"]]
    table = table.mutate(id2=table["id"])
    result = table.execute()
    tm.assert_series_equal(result["id"], result["id2"], check_names=False)


def test_interval_columns_invalid(con):
    msg = r"DayTimeIntervalType\(0, 1\) couldn't be converted to Interval"
    with pytest.raises(IbisTypeError, match=msg):
        con.table("invalid_interval_table")


def test_string_literal_backslash_escaping(con):
    expr = ibis.literal("\\d\\e")
    result = con.execute(expr)
    assert result == "\\d\\e"



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest

import ibis
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import _
from ibis.common.annotations import SignatureValidationError
from ibis.tests.util import assert_equal


@pytest.fixture
def sql_table():
    return ibis.table(
        [
            ("v1", "decimal(12, 2)"),
            ("v2", "decimal(10, 4)"),
            ("v3", "int32"),
            ("v4", "int64"),
            ("v5", "float32"),
            ("v6", "double"),
            ("v7", "string"),
            ("v8", "boolean"),
        ],
        "testing",
    )


@pytest.fixture(params=(ibis.coalesce, ibis.greatest, ibis.least))
def function(request):
    return request.param


@pytest.mark.parametrize(
    "colname",
    [
        "tinyint_col",
        "smallint_col",
        "int_col",
        "bigint_col",
        "float_col",
        "double_col",
    ],
)
def test_abs(functional_alltypes, lineitem, colname):
    fname = "abs"
    op = ops.Abs

    expr = functional_alltypes[colname]
    _check_unary_op(expr, fname, op, type(expr))

    expr = lineitem.l_extendedprice
    _check_unary_op(expr, fname, op, type(expr))


def test_group_concat(functional_alltypes):
    col = functional_alltypes.string_col

    expr = col.group_concat()
    assert isinstance(expr.op(), ops.GroupConcat)
    op = expr.op()
    assert op.sep == ibis.literal(",").op()
    assert op.where is None

    expr = col.group_concat("|")
    op = expr.op()
    assert op.sep == ibis.literal("|").op()
    assert op.where is None


def test_zero_ifnull(functional_alltypes):
    dresult = functional_alltypes.double_col.fillna(0)

    iresult = functional_alltypes.int_col.fillna(0)

    assert type(dresult.op()) == ops.Coalesce
    assert type(dresult) == ir.FloatingColumn

    # Impala upconverts all ints to bigint. Hmm.
    assert type(iresult) == type(iresult)


def test_fillna(functional_alltypes):
    result = functional_alltypes.double_col.fillna(5)
    assert isinstance(result, ir.FloatingColumn)

    assert isinstance(result.op(), ops.Coalesce)

    result = functional_alltypes.bool_col.fillna(True)
    assert isinstance(result, ir.BooleanColumn)

    # Highest precedence type
    result = functional_alltypes.int_col.fillna(functional_alltypes.bigint_col)
    assert isinstance(result, ir.IntegerColumn)


def test_ceil_floor(functional_alltypes, lineitem):
    cresult = functional_alltypes.double_col.ceil()
    fresult = functional_alltypes.double_col.floor()
    assert isinstance(cresult, ir.IntegerColumn)
    assert isinstance(fresult, ir.IntegerColumn)
    assert type(cresult.op()) == ops.Ceil
    assert type(fresult.op()) == ops.Floor

    cresult = ibis.literal(1.2345).ceil()
    fresult = ibis.literal(1.2345).floor()
    assert isinstance(cresult, ir.IntegerScalar)
    assert isinstance(fresult, ir.IntegerScalar)

    dec_col = lineitem.l_extendedprice
    cresult = dec_col.ceil()
    fresult = dec_col.floor()
    assert isinstance(cresult, ir.DecimalColumn)
    assert cresult.type() == dec_col.type()

    assert isinstance(fresult, ir.DecimalColumn)
    assert fresult.type() == dec_col.type()


def test_sign(functional_alltypes, lineitem):
    result = functional_alltypes.double_col.sign()
    assert isinstance(result, ir.FloatingColumn)
    assert type(result.op()) == ops.Sign

    result = ibis.literal(1.2345).sign()
    assert isinstance(result, ir.FloatingScalar)

    dec_col = lineitem.l_extendedprice
    result = dec_col.sign()
    assert isinstance(result, ir.DecimalColumn)


def test_round(functional_alltypes, lineitem):
    result = functional_alltypes.double_col.round()
    assert isinstance(result, ir.IntegerColumn)
    assert result.op().args[1] is None

    result = functional_alltypes.double_col.round(2)
    assert isinstance(result, ir.FloatingColumn)
    assert result.op().args[1] == ibis.literal(2).op()

    # Even integers are double (at least in Impala, check with other DB
    # implementations)
    result = functional_alltypes.int_col.round(2)
    assert isinstance(result, ir.FloatingColumn)

    dec = lineitem.l_extendedprice
    result = dec.round()
    assert isinstance(result, ir.DecimalColumn)

    result = dec.round(2)
    assert isinstance(result, ir.DecimalColumn)

    result = ibis.literal(1.2345).round()
    assert isinstance(result, ir.IntegerScalar)


def _check_unary_op(expr, fname, ex_op, ex_type):
    result = getattr(expr, fname)()
    assert type(result.op()) == ex_op
    assert type(result) == ex_type


def test_coalesce_instance_method(sql_table):
    v7 = sql_table.v7
    v5 = sql_table.v5.cast("string")
    v8 = sql_table.v8.cast("string")

    result = v7.coalesce(v5, v8, "foo")
    expected = ibis.coalesce(v7, v5, v8, "foo")
    assert_equal(result, expected)


def test_integer_promotions(sql_table, function):
    t = sql_table

    expr = function(t.v3, t.v4)
    assert isinstance(expr, ir.IntegerColumn)

    expr = function(5, t.v3)
    assert isinstance(expr, ir.IntegerColumn)

    expr = function(5, 12)
    assert isinstance(expr, ir.IntegerScalar)


def test_floats(sql_table, function):
    t = sql_table

    expr = function(t.v5)
    assert isinstance(expr, ir.FloatingColumn)

    expr = function(5.5, t.v5)
    assert isinstance(expr, ir.FloatingColumn)

    expr = function(5.5, 5)
    assert isinstance(expr, ir.FloatingScalar)


def test_deferred(sql_table, function):
    expr = function(None, _.v3, 2)
    res = expr.resolve(sql_table)
    sol = function(None, sql_table.v3, 2)
    assert res.equals(sol)


def test_no_arguments_errors(function):
    with pytest.raises(
        SignatureValidationError, match=".+ has failed due to the following errors:"
    ):
        function()
